using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

namespace IdeaBlade.EntityModel {

  #region Non-generic EntityQueryOperation Extensions

  public static class EntityQueryOperationExtensions {

    public static Func<EntityQueryOperation, IEntityQueryOperationWrapper> OperationWrapperFactory =
      op => new DefaultEntityQueryOperationWrapper(op);

    private static IEntityQueryOperationWrapper MakeWrapper(EntityQueryOperation op) {
      return OperationWrapperFactory(op);
    }

    public static IEntityQueryOperationWrapper OnCancelled (this EntityQueryOperation op, Action action) {
      return MakeWrapper(op).OnCancelled(action);
    }

    /// <summary>Equivalent to operation's Completed event; terminates the "On..." chain.</summary>
    public static void OnCompleted(this EntityQueryOperation op, Action<EntityQueriedEventArgs> action) {
      MakeWrapper(op).OnCompleted(action);
    }

    public static IEntityQueryOperationWrapper OnError(this EntityQueryOperation op, Action<Exception> action) {
      return MakeWrapper(op).OnError(action);
    }

    public static IEntityQueryOperationWrapper OnError(this EntityQueryOperation op, string failureMessageTemplate = null) {
      return MakeWrapper(op).OnError(failureMessageTemplate);
    }

    public static IEntityQueryOperationWrapper Into(this EntityQueryOperation op, IList targetCollection) {
      return MakeWrapper(op).Into(targetCollection);
    }


    public static IEntityQueryOperationWrapper OnSuccess(this EntityQueryOperation op, Action<IEnumerable> action) {
      return MakeWrapper(op).OnSuccess(action);
    }
  }

  public interface IEntityQueryOperationWrapper  {

    IEntityQueryOperationWrapper Into(IList targetCollection);
    IEntityQueryOperationWrapper OnSuccess(Action<IEnumerable> action);
    IEntityQueryOperationWrapper OnCancelled(Action action);
    /// <summary>Equivalent to operation's Completed event; terminates the "On..." chain.</summary>
    void OnCompleted(Action<EntityQueriedEventArgs> action);
    IEntityQueryOperationWrapper OnError(Action<Exception> action);
    IEntityQueryOperationWrapper OnError(string failureMessageTemplate = null);

  }

  public class DefaultEntityQueryOperationWrapper : IEntityQueryOperationWrapper {

    public DefaultEntityQueryOperationWrapper(EntityQueryOperation op) {
      Operation = op;
      OnCompletedHandler = new OperationOnCompletedHandler<EntityQueryOperation, EntityQueriedEventArgs>(op);
      op.Completed += OnCompletedHandler.CompletedHandler;
    }

    public EntityQueryOperation Operation { get; private set; }

    protected OperationOnCompletedHandler<EntityQueryOperation, EntityQueriedEventArgs> OnCompletedHandler { get; set; }

    public IEntityQueryOperationWrapper OnCancelled(Action action) {
      OnCompletedHandler.OnCancelled(action);
      return this;
    }

    public void OnCompleted(Action<EntityQueriedEventArgs> action) {
      OnCompletedHandler.OnCompleted(action);
    }

    public IEntityQueryOperationWrapper OnError(Action<Exception> action) {
      OnCompletedHandler.OnError(action);
      return this;
    }

    public IEntityQueryOperationWrapper OnError(string failureMessageTemplate = null) {
      OnCompletedHandler.OnError(MakeFailureMessageTemplate(failureMessageTemplate));
      return this;
    }

    private string MakeFailureMessageTemplate(string failureMessageTemplate) {
      if (String.IsNullOrWhiteSpace(failureMessageTemplate)) {
        failureMessageTemplate = "EntityQuery failed with exception \"{0}\"."; 
      }
      return failureMessageTemplate;
    }

    public IEntityQueryOperationWrapper Into(IList targetCollection) {
      OnCompletedHandler.OnSuccess(args => {
                                     var results = args.Results;
                                     foreach (var item in results) {
                                       targetCollection.Add(item);
                                     }
                                   });
      return this;
    }

    public IEntityQueryOperationWrapper OnSuccess(Action<IEnumerable> action) {
      OnCompletedHandler.OnSuccess( args => action(args.Results));
      return this;
    }

  }

  #endregion

  #region EntityQueryOperation<T> Extensions

  public static class EntityQueryOperationOfTExtensions {

    // ToDo: Figure out how to substitute a custom MakeWrapper implementation.
    // Hurdle is can't creat extension method in a generic static class

    private static IEntityQueryOperationWrapper<T> MakeWrapper<T>(EntityQueryOperation<T> op) {
      return new DefaultEntityQueryOperationWrapper<T>(op);
    }

    public static IEntityQueryOperationWrapper<T> OnCancelled<T>(this EntityQueryOperation<T> op, Action action) {
      return MakeWrapper(op).OnCancelled(action);
    }

    /// <summary>Equivalent to operation's Completed event; terminates the "On..." chain.</summary>
    public static void OnCompleted<T>(this EntityQueryOperation<T> op, Action<EntityQueriedEventArgs<T>> action) {
      MakeWrapper(op).OnCompleted(action);
    }

    public static IEntityQueryOperationWrapper<T> OnError<T>(this EntityQueryOperation<T> op, Action<Exception> action) {
      return MakeWrapper(op).OnError(action);
    }

    public static IEntityQueryOperationWrapper<T> OnError<T>(this EntityQueryOperation<T> op, string failureMessageTemplate = null) {
      return MakeWrapper(op).OnError(failureMessageTemplate);
    }

    public static IEntityQueryOperationWrapper<T> Into<T>(this EntityQueryOperation<T> op, IList targetCollection) {
      return MakeWrapper(op).Into(targetCollection);
    }

    public static IEntityQueryOperationWrapper<T> OnSuccess<T>(this EntityQueryOperation<T> op, Action<IEnumerable<T>> action) {
      return MakeWrapper(op).OnSuccess(action);
    }
  }

  public interface IEntityQueryOperationWrapper<T> {

    IEntityQueryOperationWrapper<T> Into(IList targetCollection);
    IEntityQueryOperationWrapper<T> OnSuccess(Action<IEnumerable<T>> action);
    IEntityQueryOperationWrapper<T> OnCancelled(Action action);
    /// <summary>Equivalent to operation's Completed event; terminates the "On..." chain.</summary>
    void OnCompleted(Action<EntityQueriedEventArgs<T>> action);
    IEntityQueryOperationWrapper<T> OnError(Action<Exception> action);
    IEntityQueryOperationWrapper<T> OnError(string failureMessageTemplate = null);

  }

  public class DefaultEntityQueryOperationWrapper<T> : IEntityQueryOperationWrapper<T> {

    public DefaultEntityQueryOperationWrapper(EntityQueryOperation<T> op) {
      Operation = op;
      OnCompletedHandler = new OperationOnCompletedHandler<EntityQueryOperation<T>, EntityQueriedEventArgs<T>>(op);
      op.Completed += OnCompletedHandler.CompletedHandler;
    }

    public EntityQueryOperation<T> Operation { get; private set; }

    protected OperationOnCompletedHandler<EntityQueryOperation<T>, EntityQueriedEventArgs<T>> OnCompletedHandler { get; set; }

    public IEntityQueryOperationWrapper<T> OnCancelled(Action action) {
      OnCompletedHandler.OnCancelled(action);
      return this;
    }

    public void OnCompleted(Action<EntityQueriedEventArgs<T>> action) {
      OnCompletedHandler.OnCompleted(action);
    }

    public IEntityQueryOperationWrapper<T> OnError(Action<Exception> action) {
      OnCompletedHandler.OnError(action);
      return this;
    }

    public IEntityQueryOperationWrapper<T> OnError(string failureMessageTemplate = null) {
      OnCompletedHandler.OnError(MakeFailureMessageTemplate(failureMessageTemplate));
      return this;
    }

    private string MakeFailureMessageTemplate(string failureMessageTemplate) {
      if (String.IsNullOrWhiteSpace(failureMessageTemplate)) {
        var queryTypeName = ((IQueryable) Operation.EntityQuery).ElementType.Name;
        failureMessageTemplate = "EntityQuery<" + queryTypeName + "> failed with exception \"{0}\".";
      }
      return failureMessageTemplate;
    }

    public IEntityQueryOperationWrapper<T> Into(IList targetCollection) {
      OnCompletedHandler.OnSuccess(args => {
                                     var results = args.Results;
                                     foreach (var item in results) {
                                       targetCollection.Add(item);
                                     }
                                   });
      return this;
    }

    public IEntityQueryOperationWrapper<T> OnSuccess(Action<IEnumerable<T>> action) {
      OnCompletedHandler.OnSuccess(args => action(args.Results));
      return this;
    }

  }

  #endregion

}