using System;

namespace IdeaBlade.EntityModel {

  public static class EntityScalarQueryOperationExtensions {

    // ToDo: Figure out how to substitute a custom MakeWrapper implementation.

    private static IEntityScalarQueryOperationWrapper<T> MakeWrapper<T>(EntityScalarQueryOperation<T> op) {
      return new DefaultEntityScalarQueryOperationWrapper<T>(op);
    }

    public static IEntityScalarQueryOperationWrapper<T> OnCancelled<T>(this EntityScalarQueryOperation<T> op, Action action) {
      return MakeWrapper(op).OnCancelled(action);
    }

    /// <summary>Equivalent to operation's Completed event; terminates the "On..." chain.</summary>
    public static void OnCompleted<T>(this EntityScalarQueryOperation<T> op, Action<EntityScalarQueriedEventArgs<T>> action) {
      MakeWrapper(op).OnCompleted(action);
    }

    public static IEntityScalarQueryOperationWrapper<T> OnError<T>(this EntityScalarQueryOperation<T> op, Action<Exception> action) {
      return MakeWrapper(op).OnError(action);
    }

    public static IEntityScalarQueryOperationWrapper<T> OnError<T>(this EntityScalarQueryOperation<T> op, string failureMessageTemplate = null) {
      return MakeWrapper(op).OnError(failureMessageTemplate);
    }

    public static IEntityScalarQueryOperationWrapper<T> OnSuccess<T>(this EntityScalarQueryOperation<T> op, Action<T> action) {
      return MakeWrapper(op).OnSuccess(action);
    }
  }

  public interface IEntityScalarQueryOperationWrapper<T> {

    IEntityScalarQueryOperationWrapper<T> OnSuccess(Action<T> action);
    IEntityScalarQueryOperationWrapper<T> OnCancelled(Action action);
    /// <summary>Equivalent to operation's Completed event; terminates the "On..." chain.</summary>
    void OnCompleted(Action<EntityScalarQueriedEventArgs<T>> action);
    IEntityScalarQueryOperationWrapper<T> OnError(Action<Exception> action);
    IEntityScalarQueryOperationWrapper<T> OnError(string failureMessageTemplate = null);

  }

  public class DefaultEntityScalarQueryOperationWrapper<T> : IEntityScalarQueryOperationWrapper<T> {

    public DefaultEntityScalarQueryOperationWrapper(EntityScalarQueryOperation<T> op) {
      Operation = op;
      OnCompletedHandler = new OperationOnCompletedHandler<EntityScalarQueryOperation<T>, EntityScalarQueriedEventArgs<T>>(op);
      op.Completed += OnCompletedHandler.CompletedHandler;
    }

    public EntityScalarQueryOperation<T> Operation { get; private set; }

    protected OperationOnCompletedHandler<EntityScalarQueryOperation<T>, EntityScalarQueriedEventArgs<T>> OnCompletedHandler { get; set; }

    public IEntityScalarQueryOperationWrapper<T> OnCancelled(Action action) {
      OnCompletedHandler.OnCancelled(action);
      return this;
    }

    public void OnCompleted(Action<EntityScalarQueriedEventArgs<T>> action) {
      OnCompletedHandler.OnCompleted(action);
    }

    public IEntityScalarQueryOperationWrapper<T> OnError(Action<Exception> action) {
      OnCompletedHandler.OnError(action);
      return this;
    }

    public IEntityScalarQueryOperationWrapper<T> OnError(string failureMessageTemplate = null) {
      OnCompletedHandler.OnError(MakeFailureMessageTemplate(failureMessageTemplate));
      return this;
    }

    private string MakeFailureMessageTemplate(string failureMessageTemplate) {
      if (String.IsNullOrWhiteSpace(failureMessageTemplate)) {
        var queryTypeName = Operation.EntityQuery.ElementType.Name;
        failureMessageTemplate = "ScalarQuery<" + queryTypeName + "> failed with exception \"{0}\".";
      }
      return failureMessageTemplate;
    }

    public IEntityScalarQueryOperationWrapper<T> OnSuccess(Action<T> action) {
      OnCompletedHandler.OnSuccess(args => action(args.Result));
      return this;
    }

  }

}