using System;
using System.ComponentModel;
// ReSharper disable InconsistentNaming

namespace IdeaBlade.EntityModel {

  public class PropertyChangedCoroutineOperation : INotifyCompleted {

    /// <summary>
    /// Wait upon a <see cref="INotifyPropertyChanged.PropertyChanged"/>
    /// event for the specified property.
    /// </summary>
    /// <param name="inpc">
    /// Object that implements <see cref="INotifyPropertyChanged.PropertyChanged"/>.
    /// </param>
    /// <param name="propertyName">
    /// Name of the property to watch. Null or empty means watch for any
    /// <see cref="INotifyPropertyChanged.PropertyChanged"/> event.
    /// </param>
    /// <param name="matchEmptyPropertyName">
    /// If true, will also complete if the PropertyName in the
    /// <see cref="INotifyPropertyChanged.PropertyChanged"/> EventArgs is
    /// null or empty.
    /// </param>
    public PropertyChangedCoroutineOperation(
      INotifyPropertyChanged inpc,
      string propertyName = null,
      bool matchEmptyPropertyName = false)
      : this(inpc, null, propertyName, matchEmptyPropertyName) { }

    /// <summary>
    /// Wait upon a <see cref="INotifyPropertyChanged.PropertyChanged"/>
    /// event for the specified property and for the property value to
    /// satisfy the <see cref="propertyValueMatchPredicate"/> test.
    /// </summary>
    /// <param name="inpc">
    /// Object that implements <see cref="INotifyPropertyChanged.PropertyChanged"/>.
    /// </param>
    /// <param name="propertyValueMatchPredicate">
    /// Function that reports true if the the operation should complete.
    /// Typically tests for a specific value of the watched property.
    /// Completes immediately if the function returns true from the start.
    /// </param>
    /// <param name="propertyName">
    /// Name of the property to watch. Null or empty means watch for any
    /// <see cref="INotifyPropertyChanged.PropertyChanged"/> event.
    /// </param>
    /// <param name="matchEmptyPropertyName">
    /// If true, will also complete if the PropertyName in the
    /// <see cref="INotifyPropertyChanged.PropertyChanged"/> EventArgs is
    /// null or empty.
    /// </param>
    /// <remarks>
    /// The <see cref="propertyValueMatchPredicate"/>, if non-null,
    /// is evaluated during construction and, if true, the operation completes
    /// immediately without bothering to wait on PropertyChanged.
    /// </remarks>
    public PropertyChangedCoroutineOperation(
      INotifyPropertyChanged inpc,
      Func<bool> propertyValueMatchPredicate,
      string propertyName = null,
      bool matchEmptyPropertyName = false) {

      CompletedEventArgs =
        new PropertyChangedCoroutineCompletedEventArgs(
          inpc, propertyValueMatchPredicate, propertyName, matchEmptyPropertyName);

      CompletedEventArgs.Completed += (s, args) => RaiseCompleted();
    }

    public string ActualPropertyName { get { return CompletedEventArgs.ActualPropertyName; } }
    public string ExpectedPropertyName { get { return CompletedEventArgs.ExpectedPropertyName; } }
    public bool IsCompleted { get { return CompletedEventArgs.IsCompleted; } }
    public bool MatchEmptyPropertyName { get { return CompletedEventArgs.MatchEmptyPropertyName; } }

    private readonly PropertyChangedCoroutineCompletedEventArgs CompletedEventArgs;

    public virtual void WhenCompleted(Action<INotifyCompletedArgs> completedAction) {
      if (null == completedAction) return;
      if (IsCompleted) {
        completedAction(CompletedEventArgs);
        return;
      }
      _notifyCompletedActions =
        (Action<INotifyCompletedArgs>) Delegate.Combine(_notifyCompletedActions, completedAction);
    }

    protected void RaiseCompleted() {
      CallCompletedActions();
      var handlers = _completed;
      _completed = null;
      if (null == handlers) return;
      handlers(this, CompletedEventArgs);
    }

    private event EventHandler<PropertyChangedCoroutineCompletedEventArgs> _completed;
    public event EventHandler<PropertyChangedCoroutineCompletedEventArgs> Completed {
      add {
        if (IsCompleted) {
          value(this, CompletedEventArgs);
          return;
        }
        _completed += value;
      }
      remove { _completed -= value; }
    }

    protected void CallCompletedActions() {
      var actions = _notifyCompletedActions;
      _notifyCompletedActions = null;
      if (null == actions) return;
      actions(CompletedEventArgs);
    }

    private Action<INotifyCompletedArgs> _notifyCompletedActions;

  }

  public class PropertyChangedCoroutineCompletedEventArgs : EventArgs, INotifyCompletedArgs {

    /// <summary>
    /// Wait upon a <see cref="INotifyPropertyChanged.PropertyChanged"/>
    /// event for the specified property and for the property value to
    /// satisfy the <see cref="propertyValueMatchPredicate"/> test.
    /// </summary>
    /// <param name="inpc">
    /// Object that implements <see cref="INotifyPropertyChanged.PropertyChanged"/>.
    /// </param>
    /// <param name="propertyValueMatchPredicate">
    /// Function that reports true if the the operation should complete.
    /// Typically tests for a specific value of the watched property.
    /// Completes immediately if the function returns true from the start.
    /// </param>
    /// <param name="propertyName">
    /// Name of the property to watch. Null or empty means watch for any
    /// <see cref="INotifyPropertyChanged.PropertyChanged"/> event.
    /// </param>
    /// <param name="matchEmptyPropertyName">
    /// If true, will also complete if the PropertyName in the
    /// <see cref="INotifyPropertyChanged.PropertyChanged"/> EventArgs is
    /// null or empty.
    /// </param>
    /// <remarks>
    /// The <see cref="propertyValueMatchPredicate"/>, if non-null,
    /// is evaluated during construction and, if true, the operation completes
    /// immediately without bothering to wait on PropertyChanged.
    /// </remarks>
    public PropertyChangedCoroutineCompletedEventArgs(
      INotifyPropertyChanged inpc,
      Func<bool> propertyValueMatchPredicate,
      string propertyName,
      bool matchEmptyPropertyName) {

      Inpc = inpc;

      ExpectedPropertyName = propertyName;
      _matchAnyPropertyName = String.IsNullOrEmpty(ExpectedPropertyName);

      MatchEmptyPropertyName = matchEmptyPropertyName;
      _matchEmptyPropertyName = matchEmptyPropertyName || String.IsNullOrEmpty(propertyName);

      PropertyValueMatchPredicate = propertyValueMatchPredicate;

      // Done if property value already matches
      if (null != propertyValueMatchPredicate && propertyValueMatchPredicate()) {
        ActualPropertyName = ExpectedPropertyName;
        RaiseCompleted();
      }

      _propertyValueMatchPredicate = propertyValueMatchPredicate ?? delegate { return true; };
      inpc.PropertyChanged += PropertyChangedHandler;
    }

    private readonly bool _matchEmptyPropertyName;
    private readonly bool _matchAnyPropertyName;
    private readonly Func<bool> _propertyValueMatchPredicate;

    private void PropertyChangedHandler(object sender, PropertyChangedEventArgs e) {
      var actualPropertyName = e.PropertyName;
      if (_matchAnyPropertyName ||
          actualPropertyName == ExpectedPropertyName ||
          (String.IsNullOrEmpty(actualPropertyName) && _matchEmptyPropertyName)) {

          if (!_propertyValueMatchPredicate()) return; // didn't match required value

          ActualPropertyName = actualPropertyName;
          Inpc.PropertyChanged -= PropertyChangedHandler; // stop listening
          RaiseCompleted();
          Inpc = null; // let go of inpc object
      }
    }

    private void RaiseCompleted() {
      IsCompleted = true;
      if (null == Completed) return;
      Completed(Inpc, this);
    }

    internal event EventHandler<PropertyChangedCoroutineCompletedEventArgs> Completed;
    public bool IsCompleted { get; private set; }
    public string ActualPropertyName { get; private set; }
    public string ExpectedPropertyName { get; private set; }
    public bool MatchEmptyPropertyName { get; private set; }
    public Func<bool> PropertyValueMatchPredicate { get; private set; }
    private INotifyPropertyChanged Inpc { get; set; }


    public bool Cancelled { get { return false; } } // can't cancel
    public Exception Error { get { return null; } } // can't fail
    public bool IsErrorHandled { get; set; } // irrelevant

  }
}