using System;
using System.Collections.Generic;
using System.Linq;
using IdeaBlade.TestFramework;
using Microsoft.Silverlight.Testing;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using SoftwareApproach.TestingExtensions;

using IdeaBlade.EntityModel;
using SimpleNorthwindModel;

// ReSharper disable InconsistentNaming

namespace Querying.AsyncQueryInclude {

  [TestClass]
  public class When_query_with_include : DevForceTest {

    [TestMethod, Asynchronous, Timeout(15000)]
    [Tag("Include")]
    public void then_can_get_Orders_with_Customers() {

      DoItAsync(() => {

        var custQueryResult = new List<Customer>();
        var manager = new NorthwindManager();
        manager.Customers
         .Take(5)
         .Include("Orders") // <-- comment out to see test fail for lack of orders
         .ExecuteAsync()
         .Completed += (sender, args) => {
           if (args.HasError) Assert.Fail("Customers query error: " + args.Error);
           custQueryResult.AddRange(args.Results);

           // Assertion: the query retrieved orders with the customers
           //   "ChangedEntities" reports entities from this particular query 
           //    that changed the EntityManager cache
           //    Contains any retrieved entity which was added to the cache or which
           //    modified an entity already in the cache.
           //    If we ran the query again for this manager,
           //    "ChangedEntities would be empty.
           args.ChangedEntities.Any(entity => entity is Order)
             .ShouldBeTrue("Did not get orders with my customers");

           TestComplete(); // so test method closes
         };

      });
    }

    [TestMethod, Asynchronous, Timeout(15000)]
    [Tag("Include")]
    public void then_can_get_Orders_with_Customers_with_no_magic_string() {

      DoItAsync(() => {

        var custQueryResult = new List<Customer>();
        var manager = new NorthwindManager();
        manager.Customers
         .Take(5)
         .Include(c => c.Orders) // <-- lambda instead of magic string "Orders"
         .ExecuteAsync()
         .Completed += (sender, args) => {
           if (args.HasError) Assert.Fail("Customers query error: " + args.Error);
           custQueryResult.AddRange(args.Results);

           args.ChangedEntities.Any(entity => entity is Order)
             .ShouldBeTrue("Did not get orders with my customers");

           TestComplete(); // so test method closes
         };

      });
    }

    // Verify that it doesn't go to the server during navigation
    // by preventing the manager from trying
    [TestMethod, Asynchronous, Timeout(15000)]
    [Tag("Include")]
    public void then_does_not_go_to_server_for_Customer_Orders() {

      DoItAsync(() => {

        var custQueryResult = new List<Customer>();
        var manager = new NorthwindManager();
        manager.Customers
         .Take(5)
         .Include(c => c.Orders) // <-- lambda instead of magic string "Orders"
         .ExecuteAsync()
         .Completed += (sender, args) => {

           if (args.HasError) Assert.Fail("Customers query error: " + args.Error);
           custQueryResult.AddRange(args.Results);

           manager.Disconnect(); // Manager should now look only in cache
           manager.Fetching += delegate { Assert.Fail("Tried to navigate to the server"); };

           Customer firstCustomer = custQueryResult.First();
           IList<Order> firstOrders = firstCustomer.Orders; 
           firstOrders.ShouldNotBeEmpty("Did not get orders with my customers");

           TestComplete(); // so test method closes
         };

      });
    }

    [TestMethod, Asynchronous, Timeout(15000)]
    [Tag("Include")]
    public void then_can_get_Orders_and_OrderDetails_of_Customers() {

      DoItAsync(() => {

        var custQueryResult = new List<Customer>();
        var manager = new NorthwindManager();
        manager.Customers
         .Take(5)
         // Can't use strong typing (lambda) if a middle leg
         // is a collection. Here "Orders" is a collection
         .Include("Orders.OrderDetails") 
         .ExecuteAsync()
         .Completed += (sender, args) => {
           if (args.HasError) Assert.Fail("Customers query error: " + args.Error);
           custQueryResult.AddRange(args.Results);

           args.ChangedEntities.Any(entity => entity is Order)
             .ShouldBeTrue("Did not get Orders with my customers");
           args.ChangedEntities.Any(entity => entity is OrderDetail)
             .ShouldBeTrue("Did not get OrderDetails with my customers");
           TestComplete(); // so test method closes
         };

      });
    }
//---------------------------------------------------
   
  }

  /// <summary>
  /// Explore asynchronous queries without "Include"
  /// Use entity navigation to determine if entities are in cache.
  /// an approach that requires preventing "Lazy load" by
  /// blocking the EntityManager from querying the database inside
  /// a navigation property.
  /// </summary>
  [TestClass]
  public class When_async_query_without_include
    : AsyncQueryIncludeContext {

    /// <remarks>
    /// Use entity navigation to determine if entities are in cache.
    /// an approach that requires preventing "Lazy load" by
    /// blocking the EntityManager from querying the database inside
    /// a navigation property.
    /// </remarks>
    [TestMethod, Asynchronous, Timeout(20000)]
    [Tag("Integration"), Tag("QueryInclude")]
    public void then_dont_get_Nancys_Orders_from_cache_by_navigation() {

      DoItAsync( () =>

          // Use FirstOrNullEntity to avoid possible NullReferenceException during navigation
          NancyAsyncQuery
            .FirstOrNullEntity()
            .TestTheResult(
              nancy => {
                ForceManagerLocal(); // prevents lazy load
                // Should get an empty list after navigating from Nancy to her Orders 
                nancy.Orders.ShouldBeEmpty("Should NOT find orders for Nancy");
              }
            )              
        );
    }

    /// <remarks>
    /// Use <see cref="EntityManager.FindEntities{T}(EntityState)"/>
    /// to determine if entities are in cache.
    /// </remarks>
    [TestMethod, Asynchronous, Timeout(20000)]
    [Tag("Integration"), Tag("QueryInclude")]
    public void then_Nancy_does_not_have_orders_in_cache() {

      DoItAsync(() =>

          // We don't care about the query result so FirstOrDefault is fine
          NancyAsyncQuery
            .FirstOrDefault()
            .TestTheResult(
              result => {
                // Use FindEntities to poke around in the EntityManager's entity cache 
                var orders = Manager.FindEntities<Order>(EntityStates.AllCachedStates);
                orders.ShouldBeEmpty("Should NOT find orders for Nancy");
              }
            )
        );
    }
  }

  //-------------------------------------------------------------------------

  /// <summary>
  /// Explore asynchronous queries with and without "Include"
  /// Use entity navigation to determine if entities are in cache.
  /// an approach that requires preventing "Lazy load" by
  /// blocking the EntityManager from querying the database inside
  /// a navigation property.
  /// </summary>
  [TestClass]
  public class When_async_query_with_include 
    : AsyncQueryIncludeContext {

    /// <remarks>
    /// Use entity navigation to determine if entities are in cache.
    /// an approach that requires preventing "Lazy load" by
    /// blocking the EntityManager from querying the database inside
    /// a navigation property.
    /// </remarks>
    [TestMethod, Asynchronous, Timeout(20000)]
    [Tag("Integration"), Tag("QueryInclude")]
    public void then_prove_can_get_Nancys_Orders_from_cache_locally() {

      DoItAsync(() =>

          // Use FirstOrNullEntity to avoid possible NullReferenceException during navigation
          NancyAsyncQueryWithInclude
            .FirstOrNullEntity()
            .TestTheResult(
              nancy => {
                ForceManagerLocal(); // prevents lazy load
                // Should get a full list after navigating from Nancy to her Orders 
                nancy.Orders.ShouldNotBeEmpty("Should have orders with Nancy's ID");

              }
            )  
        );
    }

    /// <remarks>
    /// Use <see cref="EntityManager.FindEntities{T}(EntityState)"/>
    /// to determine if entities are in cache.
    /// </remarks>
    [TestMethod, Asynchronous, Timeout(20000)]
    [Tag("Integration"), Tag("QueryInclude")]
    public void then_Nancy_has_orders_in_cache() {
      DoItAsync(() =>

          // We don't care about the query result so FirstOrDefault is fine
          NancyAsyncQueryWithInclude
            .FirstOrDefault()
            .TestTheResult(
              nancy => {
                // Use FindEntities to poke around in the EntityManager's entity cache 
                var orders = Manager.FindEntities<Order>(EntityStates.AllCachedStates);
                orders.ShouldNotBeEmpty("Should have orders with Nancy's ID");
              }
            )
        );
    }
  }

  //-------------------------------------------------------------------------

  public class AsyncQueryIncludeContext : DevForceTest {

    protected override void Context() {
      // Create manager that connects to the server automatically
      Manager = new NorthwindManager(); 
      PinManagerToSyncContextThread(Manager);
    }

    #region Helpers

    /// <summary>Get and set the EntityManager used by test methods.</summary>
    protected NorthwindManager Manager { get; set; }

    protected void ForceManagerLocal() { Manager.ForceManagerLocal(); }

    /// <summary>Employee Id known to be the ID of "Nancy Davolio"</summary>
    protected const int NancyDavolioId = 1;

    /// <summary>
    /// Get a Scalar async query regarding the employee with Nancy's Id.
    /// </summary>
    protected IEntityScalarQuery<Employee> NancyAsyncQuery {
      get {
        // Which LINQ style do you prefer?

        // Fluent syntax
        return Manager.Employees
          .Where(e => e.EmployeeID == NancyDavolioId)
          .AsScalarAsync();

        /*
              // Query syntax
              var q = from emp in Manager.Employees
                      where emp.EmployeeID == NancyDavolioId
                      select emp;
              return q.AsScalarAsync();
        */
      }

    }

    /// <summary>
    /// Get a Scalar async query regarding the employee with Nancy's Id that
    /// has an "Include" statement which 
    /// brings the related orders and order details along for the ride.
    /// </summary>
    protected IEntityScalarQuery<Employee> NancyAsyncQueryWithInclude {

      get {
        // Which LINQ style do you prefer?

        /*
              // Fluent syntax
              return Manager.Employees
                .Where(e => e.EmployeeID == NancyDavolioId)
                .Include("Orders.OrderDetails") // The "Include" makes the difference
                .AsScalarAsync();
        */

        // Query syntax
        var q = from emp in Manager.Employees
                where emp.EmployeeID == NancyDavolioId
                select emp;
        q = q.Include("Orders.OrderDetails"); // The "Include" makes the difference
        return q.AsScalarAsync();
      }
    }

    #endregion

  }

}
