using System;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.Linq;
using IdeaBlade.Core;
using IdeaBlade.EntityModel;
using IdeaBlade.TestFramework;
using Microsoft.Silverlight.Testing;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using SimpleNorthwindModel;
using SoftwareApproach.TestingExtensions;

// ReSharper disable InconsistentNaming
// ReSharper disable ConvertClosureToMethodGroup
// ReSharper disable MemberCanBeMadeStatic.Local
// ReSharper disable SuggestUseVarKeywordEvident

namespace Querying.AsyncTesting {


  /// <summary>Demonstrate async queries with real database.</summary>
  [TestClass]
  public class When_querying_asynchronously : DevForceTest {

    protected override void Context() {
      _manager = null;
    }

    /// <summary>Record async operation coordination object for debugging purposes.</summary>
    private static BaseOperation BaseOpProbe { get; set; }

    #region Manager

    private NorthwindManager Manager {
      get { return _manager ?? (_manager = new NorthwindManager()); }
    }
    private NorthwindManager _manager;

    #endregion

    #region Existence query using Count, First, and First-with-projection

    private const string EXISTING_CUSTOMER_NAME = "Simons bistro";
    private const string NONEXISTING_CUSTOMER_NAME = "FooBarBaz";

    [TestMethod, Asynchronous, Timeout(10000)]
    [Tag("ExistenceQuery")]
    public void then_can_use_Any_in_a_Sync_query_to_find_if_customer_exists() {

      DoItAsync(() => {

        // Not an async test but we structure as if it were because
        // we're running it in an async test framework

        // Ensure manager is disconnected and only runs locally
        // so we can run synchronous queries even in Silverlight
        var manager = new NorthwindManager(false /* disconnected */);
        manager.DefaultQueryStrategy = QueryStrategy.CacheOnly;
        
        // Attach a test entity that will satisfy the selection criteria
        manager.AttachEntity( 
          new Customer {
              CustomerID = Guid.NewGuid(), 
              CompanyName = EXISTING_CUSTOMER_NAME
            });

        manager.Customers
          .Any(c => c.CompanyName == EXISTING_CUSTOMER_NAME)
          //.Any(c => c.CompanyName == NONEXISTING_CUSTOMER_NAME) // to see test fail
          .ShouldBeTrue("Should find the EXISTING CUSTOMER");

        TestComplete();
                
      
      });
    }              


     // There is no scalar async "Any" command; three workarounds
    // Would be useful if for existence queries such as "Does any User have this UserName?"

    /// <remarks>Without an "any" command, we use Count and test for zero.</remarks>
    [TestMethod, Asynchronous, Timeout(15000)]
    [Tag("ExistenceQuery")]
    public void then_can_use_Count_to_find_if_customer_exists() {

        DoItAsync(() => BaseOpProbe = Manager.Customers
                        .AsScalarAsync()
                        .Count(c => c.CompanyName == EXISTING_CUSTOMER_NAME)
                        //.Count(c => c.CustomerName == NONEXISTING_CUSTOMER_NAME) // to see test fail
                        .TestTheResult(result =>
                                       result.ShouldBeGreaterThan(0, "Count of custs should be positive")));
    }



    /// <remarks>First eliminates risk of table scan for count.</remarks>
    [TestMethod, Asynchronous, Timeout(15000)]
    [Tag("ExistenceQuery")]
    public void then_can_use_First_to_find_if_customer_exists() {

        DoItAsync(() => BaseOpProbe = Manager.Customers
                        .AsScalarAsync()
                        .FirstOrDefault(c => c.CompanyName == EXISTING_CUSTOMER_NAME)
                        //.FirstOrDefault(c => c.CustomerName == NONEXISTING_CUSTOMER_NAME) // to see test fail
                        .TestTheResult(result =>
                                       result.ShouldNotBeNull("First cust should not be null")));
    }

    /// <remarks>Projection minimizes payload of the response.</remarks>
    [TestMethod, Asynchronous, Timeout(15000)]
    [Tag("ExistenceQuery")]
    public void then_can_use_Projection_to_find_if_customer_exists() {

        DoItAsync(() => BaseOpProbe = Manager.Customers
                        .Where(c => c.CompanyName == EXISTING_CUSTOMER_NAME)
                        //.Where(c => c.CustomerName == NONEXISTING_CUSTOMER_NAME) // to see test fail
                        .Select(cust => "x") // shrink result to smallest payload
                        .AsScalarAsync()
                        .FirstOrDefault()
                        .TestTheResult(result =>
                                       result.ShouldNotBeNull("First cust should not be null")));
    }

    #endregion

    #region Projections

    /// <summary>
    /// Projects into an anonymous type
    /// </summary>
    [TestMethod, Asynchronous, Timeout(15000)]
    [Tag("Projection")]
    public void then_can_project_Employee_into_anonymous_type() {

      DoItAsync(() => {

                  EmployeeProjection projectedEmp;

                  BaseOpProbe = Manager.Employees
                    .Select(emp => new { emp.EmployeeID, emp.FirstName, emp.LastName, })
                    .ExecuteAsync()
                    .TestTheResults(result => {

                                      result.ShouldNotBeEmpty("Should have employee projections");

                                      var r1 = result.First();
                                      r1.GetType().Name.ShouldContain("Anonymous", "Should be anonymous type");
                                      r1.GetType().IsPublic.ShouldBeFalse("Anonymous types are not public");

                                      var bindableResult = DynamicTypeConverter.Convert(r1);
                                      bindableResult.GetType().IsPublic.ShouldBeTrue("Should be public after conversion");

                                      // Make externally excessible by pouring into a client type
                                      projectedEmp = new EmployeeProjection {
                                                                              EmployeeId = r1.EmployeeID,
                                                                              FirstName = r1.FirstName,
                                                                              LastName = r1.LastName
                                                                            };

                                      String.IsNullOrWhiteSpace(projectedEmp.FirstName)
                                        .ShouldBeFalse("FirstName should have real value");

                                    });

                });
    }

    /// <summary>
    /// Client-side type to ease access to employee slice projection
    /// </summary>
    public class EmployeeProjection {
      public int EmployeeId { get; internal set; }
      public string FirstName { get; internal set; }
      public string LastName { get; internal set; }
    }

    /// <summary>
    /// Projecting Employee
    /// into a custom type, <see cref="EmployeeSlice"/>,
    ///  defined on both server and client.
    /// </summary>
    [TestMethod, Asynchronous, Timeout(15000)]
    [Tag("Projection")]
    public void then_can_project_Employee_into_EmployeeSlice() {


        DoItAsync(() => BaseOpProbe = Manager.Employees
                        .Select(emp => new EmployeeSlice {
                                                           EmployeeID = emp.EmployeeID, 
                                                           FirstName = emp.FirstName, 
                                                           LastName = emp.LastName,
                                                         })
                        .ExecuteAsync()
                        .TestTheResults(result => {

                                          result.ShouldNotBeEmpty("Should have EmployeeSlice objects");

                                          String.IsNullOrWhiteSpace(result.First().FirstName)
                                            .ShouldBeFalse("FirstName should have real value");

                                        }));
    }


    /// <summary>
    /// Projecting Employee and count of its orders
    /// into a custom type, <see cref="EmployeeSlice"/>,
    ///  defined on both server and client.
    /// </summary>
    [TestMethod, Asynchronous, Timeout(15000)]
    [Tag("Projection")]
    public void then_can_project_Employee_and_OrderCount_into_EmployeeSlice() {

        DoItAsync(() => BaseOpProbe = Manager.Employees
                        .Select(emp => new EmployeeSlice {
                                                           EmployeeID = emp.EmployeeID,
                                                           FirstName = emp.FirstName,
                                                           LastName = emp.LastName,
                                                           OrderCount = emp.Orders.Count(),
                                                         })
                        .ExecuteAsync()
                        .TestTheResults(result => {

                                          result.ShouldNotBeEmpty("Should have EmployeeSlice objects");

                                          result.First().OrderCount
                                            .ShouldBeGreaterThan(0, "Should have projected count of emp orders");

                                        }));
    }

    #endregion

    #region Contains (IN)

    [TestMethod, Asynchronous, Timeout(15000)]
    [Tag("AsyncTesting"), Tag("Integration")]
    public void then_can_get_employees_asynchronously_by_id_list() {

      DoItAsync(delegate {

                  var idArray = new[] { 1, 3, 5 }; // won't work with arrays 
                  var idList = idArray.ToList(); //   but works fine with a list 
        
                  var op = Manager.Employees
                    .Where(emp => idList.Contains(emp.EmployeeID))
                    .ExecuteAsync();

                  op.Completed += (s, args) => {
                                    Assert.IsFalse(args.HasError, "Query exception: " + args.Error);
                                    var emps = args.Results;
                                    emps.Count().ShouldEqual(3);
                                    TestComplete();
                                  };
                });

    }

    /// <summary>
    /// Query with 'contains', the LINQ equivalent of the SQL 'IN' operator.
    /// </summary>
    [TestMethod, Asynchronous, Timeout(15000)]
    public void then_can_get_customers_in_Europe() {

      DoItAsync(delegate {

        // Prepare a LIST (not an ARRAY) of european countries
        var countryNames = new List<string> { "UK", "France" };

        // The LINQ equivalent of IN is CONTAINS
        var query = Manager.Customers
          .Where(c => countryNames.Contains(c.Country));

        query.ExecuteAsync()
            .OnError(err => Assert.Fail("Contains ('IN') query failed: " + err))
            .OnSuccess(euroCusts => {
              euroCusts.ShouldNotBeEmpty("No euro customers");
              euroCusts.Any(c => c.Country == "USA").ShouldBeFalse();
              euroCusts.Any(c => c.Country == "UK").ShouldBeTrue();
            })
            .OnCompleted(args => TestComplete());

      });
    }

    [TestMethod, Asynchronous, Timeout(15000)]
    public void then_fails_using_an_array_in_Contains() {

      DoItAsync(delegate {

        // WRONG: don't use an *array* of european countries
        string[] countryNamesArray = new[] { "UK", "France" };

        // The LINQ equivalent of IN is CONTAINS
        var query = Manager.Customers
          .Where(c => countryNamesArray.Contains(c.Country));

        BaseOpProbe = query.ExecuteAsync()
          .ExpectException<InvalidOperationException>();

      });
    }
    #endregion

    #region String match case sensitivity

    // Technique required until DF 6.0.5
    [TestMethod, Asynchronous, Timeout(15000)]
    [Tag("AsyncTesting"), Tag("Integration")]
    public void then_can_get_employees_with_case_insensitive_EQUALS_method_query() {

      DoItAsync(delegate {

                  const string lastNameToMatch = "davolio";

                  var op = Manager.Employees
                    .Where(emp => 
                           emp.LastName.Equals(lastNameToMatch, StringComparison.CurrentCultureIgnoreCase))
                    .AsScalarAsync()
                    .FirstOrDefault();

                  op.Completed += (s, args) => {
                                    Assert.IsFalse(args.HasError, "Query exception: " + args.Error);
                                    var nancy = args.Result;
                                    nancy.ShouldNotBeNull("No employee matching " + lastNameToMatch);
                                    nancy.LastName.ShouldEqual("Davolio", "Last name was not 'Davolio'");
                                    nancy.LastName.ShouldNotEqual(lastNameToMatch, "Supposed to be testing case-insensitivity");
                                    TestComplete();
                                  };
                });

    }

    // Implied case-insensitivity beginning with DF 6.0.5
    [TestMethod, Asynchronous, Timeout(15000)]
    [Tag("AsyncTesting"), Tag("Integration")]
    public void then_can_get_employees_with_case_insensitive_query() {

      DoItAsync(delegate {

                  const string lastNameToMatch = "davolio";

                  var op = Manager.Employees
                    .Where(emp => emp.LastName == lastNameToMatch)
                    .AsScalarAsync()
                    .FirstOrDefault();

                  op.Completed += (s, args) => {
                                    Assert.IsFalse(args.HasError, "Query exception: " + args.Error);
                                    var nancy = args.Result;
                                    nancy.ShouldNotBeNull("No employee matching "+lastNameToMatch);
                                    nancy.LastName.ShouldEqual("Davolio", "Last name was not 'Davolio'");
                                    nancy.LastName.ShouldNotEqual(lastNameToMatch, "Supposed to be testing case-insensitivity");
                                    TestComplete();
                                  };

                });
    }

    #endregion

    #region Exclude Entities Returned to the Client

    /// <summary>
    /// <see cref="EntityManager.Queried"/> event handler looks for 
    /// non-Brazilian orders returned after a data source fetch
    /// and removes them from the cache before the client can see or use them.
    /// Does so without disturbing the query cache which means that
    /// previously cached queries will continue to return entities from
    /// the entity cache w/o going to the server (unless told to do so).
    /// </summary>
    private void BrazilOnlyQueriedHandler(object sender, EntityQueriedEventArgs e) {
      // Remove retrieved non-Brazilian orders from entity cache
      // without disturbing the query cache.
      Manager.RemoveEntities(
      e.ChangedEntities.OfType<Order>()
        .Where(o => o.ShipCountry != "Brazil"),
        false /* don't clear query cache */
        );
    }
    
    /// <summary>
    /// Placeholder to preserve structure of all tests;
    /// you would not add this handler in real world.
    /// </summary>
    private void DoNothingQueriedHandler(object sender, EntityQueriedEventArgs e) {}

    [TestMethod, Asynchronous, Timeout(15000)]
    public void then_have_queried_Orders_from_outside_Brazil_when_no_QueriedHandler() {

      DoItAsync(() => {

        Manager.Queried += DoNothingQueriedHandler;

        Manager.Orders.ExecuteAsync()
          .OnError(ex => Assert.Fail("Order query failed with: "+ex))
          .OnSuccess(args => {
                       var ordersInCache = Manager
                         .FindEntities(typeof (Order), EntityStates.AllCachedStates)
                         .OfType<Order>();
                       var nonBrazilOrders = ordersInCache.Where(o => o.ShipCountry != "Brazil");
                       nonBrazilOrders.ShouldNotBeEmpty("Should have non-Brazilian orders");
                     })
          .OnCompleted(args => TestComplete());

        });
    }

    [TestMethod, Asynchronous, Timeout(15000)]
    public void then_no_queried_Orders_from_outside_Brazil_with_QueriedHandler() {

      DoItAsync(() => {

        Manager.Queried += BrazilOnlyQueriedHandler;

        Manager.Orders.ExecuteAsync()
          .OnError(ex => Assert.Fail("Order query failed with: " + ex))
          .OnSuccess(args => {
            var ordersInCache = Manager
              .FindEntities(typeof(Order), EntityStates.AllCachedStates)
              .OfType<Order>();
            var nonBrazilOrders = ordersInCache.Where(o => o.ShipCountry != "Brazil");
            nonBrazilOrders.ShouldBeEmpty("Should only have Brazilian orders");
          })
          .OnCompleted(args => TestComplete());
      });
    }

    [TestMethod, Asynchronous, Timeout(15000)]
    public void then_have_included_Orders_from_outside_Brazil_when_no_QueriedHandler() {

      DoItAsync(() => {

        Manager.Queried += DoNothingQueriedHandler;

        Manager.Employees
          .Include(Employee.EntityPropertyNames.Orders).ExecuteAsync()
          .OnError(ex => Assert.Fail("Order query failed with: " + ex))
          .OnSuccess(args => {
            var ordersInCache = Manager
              .FindEntities(typeof(Order), EntityStates.AllCachedStates)
              .OfType<Order>();
            var nonBrazilOrders = ordersInCache.Where(o => o.ShipCountry != "Brazil");
            nonBrazilOrders.ShouldNotBeEmpty("Should have non-Brazilian orders");
          })
          .OnCompleted(args => TestComplete());

      });
    }

    [TestMethod, Asynchronous, Timeout(20000)]
    public void then_entity_removing_queried_handler_preserves_query_caching() {

      DoItAsync(() => {

        Manager.Queried += EntityRemovingQueriedHandler;

        var query = Manager.Orders;

        query.ExecuteAsync()
          .OnError(ex => Assert.Fail("Order query failed with: " + ex))
          .OnSuccess(result => {

              // Prove that QueryHandler did remove all incoming entities
              var cachedEntities =
                Manager.FindEntities(EntityStates.AllCachedStates).OfType<object>();
              cachedEntities.ShouldBeEmpty(
                "Queried event handler should have removed everything");

              // Assert failure if this manager ever fetches from remote store again
              Manager.Fetching += delegate { Assert.Fail("Should not re-fetch"); };

              // Repeat query - which should be remembered and execute CacheOnly
              query.ExecuteAsync()
                // if no failures, the repeated query must have succeeded CacheOnly
                .OnError(ex => Assert.Fail("Re-query failed: " + ex))
                .OnCompleted(r => TestComplete());

          });

      });
    }

    private void EntityRemovingQueriedHandler(object sender, EntityQueriedEventArgs e) {
      // Removes ALL entities that would be affected by this query
      Manager.RemoveEntities(
        e.ChangedEntities,
        // This flag is CRITICAL - test fails if 'true' 
        false /* don't clear query cache */
        );
    }
    #endregion

    #region Miscellaneous
// ReSharper disable ConvertToLambdaExpression

    [TestMethod, Asynchronous, Timeout(20000)]
    public void then_async_nav_to_customer_of_order_does_not_change_order() {

      DoItAsync(() => {

#if !SILVERLIGHT
                  Manager.UseAsyncNavigation = true; // for this test
#endif
                  Manager
                    .Orders.AsScalarAsync().First()
                    .OnError(err => Assert.Fail("Order query failed: " + err))
                    .OnSuccess(order => {
                                 order.CustomerID.ShouldNotBeNull("order.CustomerID should not be null after fetch");
                                 var cust = order.Customer;
                                 cust.EntityAspect.IsPendingEntity.ShouldBeTrue(
                                   "order.Customer is not a pending entity.");
                                 order.CustomerID.ShouldNotBeNull("CustomerID should not be null after customer navigation");
                                 order.ShouldNotBeChanged("; pending customer modified the order");
                                 cust.EntityAspect.PendingEntityResolved += (s, args) => {
                                                                              args.ResolvedEntity.ShouldNotBeSameAs(cust,"Resolved cust entity should be different than cust");
                                                                              order.CustomerID.ShouldNotBeNull("CustomerID should not be null after resolution");
                                                                              order.ShouldNotBeChanged("; pending customer modified the order");
                                                                              TestComplete();
                                                                            };

                               });
                });
    }

    [TestMethod, Asynchronous, Timeout(20000)]
    [Tag("Pending")]
    public void then_no_async_nav_to_customer_of_order_within_NORMAL_query() {

      DoItAsync(() => {

#if !SILVERLIGHT
        Manager.UseAsyncNavigation = true; // for this test
#endif
        Manager
          .Orders.AsScalarAsync().First()
          .OnError(err => Assert.Fail("Order query failed: " + err))
          .OnSuccess(order => {
                       var custId = order.CustomerID;
                       custId.ShouldNotBeNull("order.CustomerID should not be null after fetch");
                       Manager.Orders.Where(o => o.Customer.CustomerID == custId)
                         .With(QueryStrategy.Normal) // <--- NORMAL STRATEGY
                         .ExecuteAsync()
                         .OnError(err => Assert.Fail("Orders-of-CustId failed: " + err))
                         .OnSuccess(results => {
                                      results.ShouldNotBeEmpty("No orders-of-CustId");
                                      var cust = results.First().Customer ?? Manager.GetNullEntity<Customer>();
                                      cust.EntityAspect.IsNullOrPendingEntity.ShouldBeFalse(
                                        "cust should be a real entity, returned by query inversion.");
                                      TestComplete();

                                    });


                     });
      });
    }

    [TestMethod, Asynchronous, Timeout(20000)]
    [Tag("Pending")]
    public void then_no_async_nav_to_customer_of_order_within_CACHEONLY_query() {

      DoItAsync(() => {

         Order firstOrder;
#if !SILVERLIGHT
         Manager.UseAsyncNavigation = true; // for this test
#endif
         Manager
           .Orders.AsScalarAsync().First()
           .OnError(err => Assert.Fail("Order query failed: " + err))
           .OnSuccess(order => {

                        firstOrder = order;
                        var custId = firstOrder.CustomerID;
                        custId.ShouldNotBeNull("order.CustomerID should not be null after fetch");

                        // Query for Customer.CustomerID
                        Manager.Orders.Where(o => o.Customer.CustomerID == custId)
                          .With(QueryStrategy.CacheOnly) // <--- CACHEONLY STRATEGY
                          .ExecuteAsync()
                          .OnError(err => Assert.Fail("Orders-of-Customer-with-CustId failed: " + err))
                          .OnSuccess(results => {

                                       // No customers in cache, therefore o.Customer is always NullEntity
                                       results.ShouldBeEmpty("Should have no Orders-of-Customer-with-CustId");

                                       // Now query for CustomerID (instead of Customer.CustomerID)
                                       Manager.Orders.Where(o => o.CustomerID == custId)
                                         .With(QueryStrategy.CacheOnly) // <--- CACHEONLY STRATEGY
                                         .ExecuteAsync()
                                         .OnError(err => Assert.Fail("Orders-with-CustId failed: " + err))
                                         .OnSuccess(results2 => {
                                             results2.Count().ShouldEqual(1, "Only one order should be in cache.");
                                             results2.ShouldContain(firstOrder, "'firstOrder' should be the only order");
                                             TestComplete();
                                           });
       
                                     });


                      });
       });
    }

// ReSharper restore ConvertToLambdaExpression
    #endregion

    #region Poco

    [TestMethod, Asynchronous, Timeout(15000)]
    public void then_can_query_Poco_in_cache() {

      DoItAsync(delegate {

           var manager = new EntityManager(false /* disconnected */); 

           var query =
             new EntityQuery<DummyEntity>()
               // must be CacheOnly because no backing store 
               // knows about this Poco type
               .With(QueryStrategy.CacheOnly);

// ReSharper restore ConvertToLambdaExpression
           manager
             .ExecuteQueryAsync(query)
             .Completed+=((s,args) => {
                            Assert.IsFalse(args.HasError, "Query failed with " + args.Error);
                            TestComplete();
                          });

         });
    }

    public class DummyEntity {
      [Key]
      public int Id { get; set; }
    }

    #endregion

  }

}