using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.Linq;
using Microsoft.Silverlight.Testing;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using SoftwareApproach.TestingExtensions;

using IdeaBlade.Core;
using IdeaBlade.EntityModel;
using SimpleNorthwindModel;
using IdeaBlade.TestFramework;

// ReSharper disable InconsistentNaming
// ReSharper disable SuggestUseVarKeywordEvident
// ReSharper disable PartialTypeWithSinglePart
// ReSharper disable RedundantAssignment

#pragma warning disable 168 // unused variable


namespace Querying.CacheOnly {

  [TestClass]
  public partial class When_querying_cache_only
    : CacheOnlyContext {

    /// <summary>Illustrate variety of ways to execute a query synchronously</summary>
    [TestMethod]
    public void then_can_get_all_employees() {

      // Define a query to get all employees
      var query = Manager.Employees;

      // Execute it with "ToList()"
      List<Employee> empsList = query.ToList();
      empsList.ShouldNotBeEmpty("Should have employees");

      // Same collection query run with variety of alternative immediate execution commands.

      IEnumerable<Employee> empsEnumerable = query.Execute();
      empsEnumerable.ShouldNotBeEmpty("Should have employees after Execute");

      Employee[] empsArray = query.ToArray();
      empsArray.ShouldNotBeEmpty("Should have employees after ToArray");

      IDictionary<int, Employee> empDictionary = query.ToDictionary(emp => emp.EmployeeID);
      empDictionary.Any().ShouldBeTrue("Should have employees after ToDictionary");

    }

    [TestMethod]
    public void then_can_get_customers_in_Europe() {

      // Prepare a LIST (not an ARRAY) of european countries
      // DevForce deliberately throws an informative exception if you try an array.
      var countryNames = new List<string> {"UK", "France"};

      // The LINQ equivalent of IN is CONTAINS
      var query = Manager.Customers
        .Where(c => countryNames.Contains(c.Country));

      var euroCusts = query.ToList();

      euroCusts.ShouldNotBeEmpty("No euro customers");
      euroCusts.Any(c => c.Country == "USA").ShouldBeFalse();
      euroCusts.Any(c => c.Country == "UK").ShouldBeTrue();

    }

    /// <summary>Illustrate variety of ways to execute a query synchronously</summary>
    [TestMethod]
    public void then_can_get_all_companies_in_a_particular_country() {

      // Query for customers in the USA - fluent syntax
      var usaQuery = Manager.Customers.Where(c => c.Country == "USA");

      // Query for customers in the UK - query syntax
      var ukQuery = from uk in Manager.Customers
                    where uk.Country == "UK"
                    select uk;
 
      // Execute each query and confirm they return something
      var usaCusts = usaQuery.ToList();
      usaCusts.ShouldNotBeEmpty("No customers in the US");

      var ukCusts = ukQuery.ToList();
      ukCusts.ShouldNotBeEmpty("No customers in the UK");

    }

    [TestMethod]
    public void then_can_find_employee_in_cache_by_entity_key() {

      // Find by EntityKey, no matter what its EntityState
      Employee emp = (Employee) Manager
        .FindEntity(new EntityKey(typeof (Employee), TestEmployeeId));

      emp.ShouldNotBeNull("Test employee not found");
      emp.EmployeeID.ShouldEqual(TestEmployeeId);

    }

    [TestMethod]
    public void then_employee_not_found_in_cache_is_null() {
      
      // When not found returns null
      var emp = (Employee) Manager
        .FindEntity(new EntityKey(typeof(Employee), int.MaxValue));

      emp.ShouldBeNull();
    }

    [TestMethod]
    public void then_can_query_for_employee_in_cache_by_id() {

      // Query by id of entity known to be in cache 
      var query = Manager.Employees
        .Where(emp => emp.EmployeeID == TestEmployeeId)
        .With(QueryStrategy.CacheOnly);

      var employee = query.SingleOrDefault();

      employee.ShouldNotBeNullOrNullEntity("Test employee not found");
      employee.EmployeeID.ShouldEqual(TestEmployeeId);

    }

    [TestMethod]
    public void then_an_employee_not_found_in_cache_by_id_query_is_nullentity() {

      // Query by id we know is NOT in cache or in the database 
      var query = Manager.Employees
        .Where(emp => emp.EmployeeID == int.MaxValue)
        .With(QueryStrategy.CacheOnly);

      var employee = query.SingleOrDefault();

      employee.ShouldBeNullOrNullEntity("employee with 'impossible' id was found");
    }

    [TestMethod]
    public void then_can_create_query_without_a_manager() {

      // Create a query to get all Employees without referring to a manager
      EntityQuery<Employee> query = new EntityQuery<Employee>();

      // Prove it works when give a manager
      query.EntityManager.ShouldBeNull("query shouldn't have a manager");
      var emps = Manager.ExecuteQuery(query);
      emps.ShouldNotBeEmpty("Should have employees");

    }

    [TestMethod]
    public void then_can_reuse_query_with_different_managers() {

      // Create a query to get all Employees without referring to a manager
      EntityQuery<Employee> query = new EntityQuery<Employee>();

      // Execute with first manager
      var emps = Manager.ExecuteQuery(query);
      emps.ShouldNotBeEmpty("Should have employees");


      // Execute with second manager (using "With" verb)
      var manager2 = PrepareNewOfflineManager();
      var emps2 = query.With(manager2);
      emps2.ShouldNotBeEmpty("Should have employees");

      // Managers are different
      Manager.ShouldNotBeSameAs(manager2);

      // Employee instances are different
      // because the managers and caches are different
      emps.ShouldNotContainAnyOf(emps2,
           "No Employee in second manager should be among first manager's employees");

      // But the employess are the same
      // from the perspective of their identities
      emps.Select(e => e.EmployeeID)
        .ShouldContainAllOf(emps.Select(e2 => e2.EmployeeID),
           "Employee IDs in second manager should be among first manager's employee IDs");
    }


    [TestMethod]
    public void then_can_import_remove_and_reimport_entities()
    {
        var cust = Manager.Customers.Where(c => c.Orders.Any()).First();
        var orders = cust.Orders;
        var q = cust.ToQuery();
        var em2 = new NorthwindManager(false /* offline */);

        em2.ImportEntities(orders.ToList().Cast<Object>().Concat(cust), MergeStrategy.OverwriteChanges);
        var sameCust = q.With(em2).First();
        Assert.IsTrue(sameCust.EntityAspect.EntityState.IsUnchanged());

        em2.RemoveEntities(typeof(Order), EntityState.AllButDetached);
        em2.RemoveEntities(typeof(Customer), EntityState.AllButDetached);
        Assert.IsTrue(sameCust.EntityAspect.EntityState.IsDetached());

        em2.ImportEntities(orders.ToList().Cast<Object>().Concat(cust), MergeStrategy.OverwriteChanges);
        var cust2 = q.With(em2).First();
        Assert.IsTrue(cust2.EntityAspect.EntityState.IsUnchanged());
    }


  }
 
  /// <summary>
  /// Demonstrates a custom filter navigation that returns
  /// a live collection maintained by the EntityManager
  /// </summary>
  [TestClass]
  public class WhenUsingFilteredNavigation : CacheOnlyContext
  {
      #region Test Initialization

      protected override void Context()
      {
          _cyOrdersChange = null;
          CurrentYearOrders = TestEmployee.CurrentYearOrders;
          CurrentYearOrders.CollectionChanged += (_, e) => _cyOrdersChange = e.Action;
      }

      NotifyCollectionChangedAction? _cyOrdersChange;
      private ObservableCollection<Order> CurrentYearOrders { get; set; }

      #endregion

      // The way it works for a generated navigation property
      [TestMethod]
      [Tag("Model"), Tag("CurrentYearOrders")]
      public void OrdersShouldChangeWhenOrderAttachedToManager()
      {
          var orders = TestEmployee.Orders;
          NotifyCollectionChangedAction? ordersChange = null;
          orders.CollectionChanged += (_, e) => ordersChange = e.Action;

          var attachedOrder =
              new Order { OrderID = 100, EmployeeID = TestEmployee.EmployeeID };

          Manager.AttachEntity(attachedOrder);

          Assert.IsTrue(orders.Contains(attachedOrder),
                        "attachedOrder not among Orders");

          Assert.IsTrue(ordersChange.HasValue, "orders did not raise change event when attaching order");
          Assert.AreEqual(NotifyCollectionChangedAction.Add, ordersChange.Value, 
              "orders raised unexpected change action when attaching order");

      }

      [TestMethod]
      [Tag("Model"), Tag("CurrentYearOrders")]
      public void ShouldHaveCurrentYearOrders()
      {
          var Orders = TestEmployee.Orders;

          Assert.AreNotEqual(0, CurrentYearOrders.Count, "CurrentYearOrders is empty");

          Assert.IsTrue(CurrentYearOrders.Count < Orders.Count,
                        "CurrentYearOrders count >= Orders");

          Assert.IsFalse(CurrentYearOrders.Any(c => c.OrderDate == null || c.OrderDate.Value.Year != 1998),
                         "Some CurrentYearOrders were not ordered in 1998");
      }

      [TestMethod]
      [Tag("Model"), Tag("CurrentYearOrders")]
      public void ShouldFindCyOrderAttachedToManager()
      {
          var newCyOrder =
              new Order { OrderID = 100, EmployeeID = TestEmployeeId, OrderDate = new DateTime(1998, 5, 1) };

          Manager.AttachEntity(newCyOrder);

          Assert.IsTrue(CurrentYearOrders.Contains(newCyOrder),
                        "newCyOrder attached to manager not among CurrentYearOrders");

          AssertCurrentYearOrdersChanged(NotifyCollectionChangedAction.Add, "order attached to manager");

      }


      [TestMethod]
      [Tag("Model"), Tag("CurrentYearOrders")]
      public void ShouldFindCYOrderAddedToCurrentYearOrders()
      {
          var newCyOrder =
              new Order { OrderID = 100, EmployeeID = TestEmployeeId, OrderDate = new DateTime(1998, 4, 1) };

          Manager.AttachEntity(newCyOrder);

          CurrentYearOrders.Add(newCyOrder);

          Assert.IsTrue(TestEmployee.CurrentYearOrders.Contains(newCyOrder),
                        "newCyOrder added to CurrentYearOrders not among subsequent CurrentYearOrders");

          AssertCurrentYearOrdersChanged(NotifyCollectionChangedAction.Add, 
              "order added to CurrentYearOrders");

      }

      [TestMethod]
      [Tag("Model"), Tag("CurrentYearOrders")]
      //[Bug("1661")] We decided not to fix this
      public void WillFindOrderAddedToCurrentYearOrdersIfNotCY()
      {
          var notCyOrder =
             new Order
             {
                 OrderID = 100,
                 EmployeeID = TestEmployee.EmployeeID + 1, // different employee
                 OrderDate = new DateTime(1996, 1, 1) // order date is not CY==1998
             }; 
          Manager.AttachEntity(notCyOrder);

          Assert.IsFalse(CurrentYearOrders.Contains(notCyOrder),
                         "Found notCyOrder among CurrentYearOrders before adding to CYOrders");

          CurrentYearOrders.Add(notCyOrder);

          //Assert.IsFalse(CurrentYearOrders.Contains(notCyOrder),// if we fixed it
          Assert.IsTrue(CurrentYearOrders.Contains(notCyOrder),// because we wont
                         "Found notCyOrder among CurrentYearOrders after adding to CYOrders"); ;

          // if we fixed it
          //AssertCurrentYearOrdersDidNotChange("1996 order for different employee added.");
      }

      // Simplified version submitted as defect
      [TestMethod]
      [Tag("Model"), Tag("CurrentYearOrders")]
      //[Bug("1661")] We decided not to fix this
      public void WillFindOrderAddedToCurrentYearOrdersIfNotCY2()
      {
          var currentYearOrders = TestEmployee.CurrentYearOrders;

          var notCyOrder =
              new Order
              {
                  OrderID = 100, 
                  EmployeeID = TestEmployee.EmployeeID + 1, // different employee
                  OrderDate = new DateTime(1996, 1, 1) // order date is not CY==1998
              };

          Manager.AttachEntity(notCyOrder);

          Assert.IsFalse(currentYearOrders.Contains(notCyOrder),
                         "Found notCyOrder among CurrentYearOrders before adding to CYOrders");

          currentYearOrders.Add(notCyOrder);

          //Assert.IsFalse(currentYearOrders.Contains(notCyOrder), // if we fixed it
          Assert.IsTrue(currentYearOrders.Contains(notCyOrder),  // because we wont
                         "Found notCyOrder among CurrentYearOrders after adding to CYOrders");

      }

      [TestMethod]
      [Tag("Model"), Tag("CurrentYearOrders")]
      public void ShouldFindOrderChangedTo1998AmongCurrentYearOrders()
      {
          var modOrder = TestEmployee.Orders.First(o => o.OrderDate == null);

          Assert.IsFalse(CurrentYearOrders.Contains(modOrder),
              "modOrder has a null OrderDate but is among CurrentYearOrders");

          modOrder.OrderDate = new DateTime(1998, 10, 4);

          Assert.IsTrue(TestEmployee.CurrentYearOrders.Contains(modOrder),
                        "modOrder is not among subsequent CurrentYearOrders after OrderDate set to 1998");

          AssertCurrentYearOrdersChanged(NotifyCollectionChangedAction.Add, 
              "order with null OrderDate set to 1998");

      }

      [TestMethod]
      [Tag("Model"), Tag("CurrentYearOrders")]
      public void ShouldNotFindOrderChangedTo1996AmongCurrentYearOrders()
      {
          var modOrder = CurrentYearOrders.First();

          modOrder.OrderDate = new DateTime(1996, 10, 4);

          Assert.IsFalse(TestEmployee.CurrentYearOrders.Contains(modOrder),
                        "modOrder is among subsequent CurrentYearOrders after OrderDate set to 1996");

          AssertCurrentYearOrdersChanged(NotifyCollectionChangedAction.Remove,
              "existing CY Order's OrderDate set to 1996");
      }

      #region Private asserts

      private void AssertCurrentYearOrdersChanged(
          NotifyCollectionChangedAction expectedChange, string whenMessage)
      {
          if (_cyOrdersChange.HasValue)
          {
              if (expectedChange == _cyOrdersChange) return;
              Assert.Fail(String.Format(
                  "CurrentYearOrders change action was {0} instead of {1} when {2}.",
                  _cyOrdersChange,
                  expectedChange,
                  whenMessage));
          }
          else
          {
              Assert.Fail("CurrentYearOrders did not raise change event when " + whenMessage);
          }
      }

      private void AssertCurrentYearOrdersDidNotChange(string whenMessage)
      {
          if (!_cyOrdersChange.HasValue) return;

          Assert.Fail(String.Format(
              "CurrentYearOrders unexpectedly changed when {1}; action was {0}.",
              _cyOrdersChange,
              whenMessage));
      }

      #endregion

  }
  
  [TestClass]
  public class WhenModifyingEntities : CacheOnlyContext
  {

#if SILVERLIGHT
        // INotifyDataErrorInfo only defined in Silverlight
        [TestMethod]
        [Tag("Model"), Tag("Mods")]
        public void ShouldNotHaveErrorsAfterRejectChanges()
        {
            var indei = (System.ComponentModel.INotifyDataErrorInfo) TestCustomer;
            Assert.IsFalse(indei.HasErrors, "Should not have errors at the start");

            TestCustomer.CompanyName = null;
            Assert.IsTrue(indei.HasErrors, "Should have error after nulling name");

            TestCustomer.EntityAspect.RejectChanges();
            TestCustomer.EntityAspect.ValidationErrors.Clear(); // ToDo: DevForce should do this?
            Assert.IsFalse(indei.HasErrors, "Should not have errors after reject changes");
        }

        [TestMethod]
        [Tag("Model"), Tag("Mods")]
        [Bug("#1663", Fixed = true )]
        public void ShouldRaiseEventWhenClearValidationErrors()
        {

            // Arrange (w/ some pre-lim asserts
            var indei = (System.ComponentModel.INotifyDataErrorInfo) TestCustomer;
            Assert.IsFalse(indei.HasErrors, "Should not have errors at the start");

            TestCustomer.CompanyName = null;
            Assert.IsTrue(indei.HasErrors, "Should have error after nulling name");

            var notified = false;
            indei.ErrorsChanged += (s,e) => notified = true;

            // Act
            TestCustomer.EntityAspect.ValidationErrors.Clear(); 

            // Assert
            Assert.IsFalse(indei.HasErrors, "Should not have errors after clearing them");
            Assert.IsTrue(notified, 
                "ErrorsChanged not called after VE.Clear().");
        }
#else
      [TestMethod]
      [Tag("Model"), Tag("Mods")]
      public void ShouldNotHaveErrorsAfterRejectChanges()
      {
          var hasError = HasError(TestCustomer, "CompanyName");
          Assert.IsFalse(hasError(), "Should not have errors at the start");

          TestCustomer.CompanyName = null;
          Assert.IsTrue(hasError(), "Should have error after nulling name");

          TestCustomer.EntityAspect.RejectChanges();
          TestCustomer.EntityAspect.ValidationErrors.Clear(); // ToDo: DevForce should do this?
          Assert.IsFalse(hasError(), "Should not have errors after reject changes");
      }
      private static Func<bool>HasError(Entity entity, string propertyName)
      {
          var idei = (System.ComponentModel.IDataErrorInfo) entity;
          return () => !String.IsNullOrEmpty(idei[propertyName]);
      }
#endif
  }                                                                                             

  [TestClass]
  public class CacheOnlyContext {
  // Silverlight Test Framework won't let us make abstract (as we might wish) 

    [TestInitialize]
    public void TestInitialize() {
      Manager = PrepareNewOfflineManager();
      Context();
    }

    protected virtual void Context() { }


    // Create manager that is disconnected and primed with test entities
    protected NorthwindManager PrepareNewOfflineManager() {

      // don't connect. Do everything in cache
      var manager = new NorthwindManager(false);
      manager.BlockFetching();

      // create some test entities
      TestCustomer =
        new Customer { CustomerID = TestCustomerId, CompanyName = "Test Company", Country = "USA" };
      TestEmployee =
        new Employee { EmployeeID = TestEmployeeId, FirstName = "Joe", LastName = "Blow" };

      // attach to the manager as if they previously existed and are unmodified.
      manager.AttachEntity(TestCustomer);
      manager.AttachEntity(TestEmployee);
      manager.AttachEntities(MakeOrders(10, TestCustomerId, TestEmployeeId));
      manager.AttachEntities(MakeMoreCustomers());

      return manager;
    }

    protected Customer TestCustomer { get; set; }
    protected Employee TestEmployee { get; set; }

    private static IEnumerable<Order> MakeOrders( int howMany, Guid? custId, int empId )
    {
        var orderDates = new DateTime?[ ]
                         {
                             new DateTime(1996, 1, 1), 
                             new DateTime(1997, 1, 1), 
                             new DateTime(1998, 1, 1), 
                             null,
                         };

        var orderID = 1;
        while (howMany-- > 0)
        {
            yield return
                new Order
                {
                    OrderID = orderID++,
                    CustomerID = custId,
                    EmployeeID = empId,
                    OrderDate = orderDates[howMany % 4],
                };
        }
    }

    private static IEnumerable<Customer> MakeMoreCustomers()
    {
        yield return new Customer {CustomerID = Guid.NewGuid(), CompanyName = "UK1 Test Company", Country = "UK"};
        yield return new Customer {CustomerID = Guid.NewGuid(), CompanyName = "UK2 Test Company", Country = "UK"};
        yield return new Customer {CustomerID = Guid.NewGuid(), CompanyName = "French1 Test Company", Country = "France"};
        yield return new Customer {CustomerID = Guid.NewGuid(), CompanyName = "French2 Test Company", Country = "France"};
        yield return new Customer {CustomerID = Guid.NewGuid(), CompanyName = "USA2 Test Company", Country = "USA"};
        yield return new Customer {CustomerID = Guid.NewGuid(), CompanyName = "USA3 Test Company", Country = "USA"};
    }

    protected NorthwindManager Manager { get; set; }
    protected static readonly Guid TestCustomerId = Guid.NewGuid();
    protected const int TestEmployeeId = 42;

    protected void ShouldGetCustomersUsingQuery(IEntityQuery<Customer> query) {
      var custs = Manager.ExecuteQuery(query);
      custs.ShouldNotBeEmpty("Can't find customers in cache");
    }

    protected void ShouldGetEmployeesUsingQuery(IEntityQuery<Employee> query) {
      var emps = Manager.ExecuteQuery(query);
      emps.ShouldNotBeEmpty("Can't find Employees in cache");
    }

  }
}
