﻿using System;
using System.Collections.Generic;
using System.Linq;
using IdeaBlade.Core;
using Microsoft.Silverlight.Testing;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using SoftwareApproach.TestingExtensions; // for Silverlight

using IdeaBlade.EntityModel;
using IdeaBlade.TestFramework;
using SimpleNorthwindModel;

// ReSharper disable InconsistentNaming
// ReSharper disable UseObjectOrCollectionInitializer

namespace Querying.CoroutineTests
{

    [TestClass]
    public class When_using_function_lists : DevForceTest
    {

        /// <summary>Executed before every test method.</summary>
        protected override void Context()
        {
            // Create manager that connects to the server automatically
            Manager = new NorthwindManager();
            PinManagerToSyncContextThread(Manager);

        }

        private NorthwindManager Manager { get; set; }

        // Just prove we can access the database
        [TestMethod, Asynchronous, Timeout(20000)]
        [Tag("Integration"), Tag("Coroutine")]
        public void can_get_customers()
        {
            DoItAsync(() =>

              Manager
                .Customers
                .ExecuteAsync()
                .TestTheResults(
                  custs => custs.ShouldNotBeEmpty("Should have customers"),
                  "Customers query")
             );
        }

        [TestMethod, Asynchronous, Timeout(20000)]
        [Tag("Integration"), Tag("Coroutine")]
        public void can_update_customers_with_serial_fnc_list()
        {
            DoItAsync(() => 
            {
                var funcList = MakeSerialAsyncFunctionList();
                Coroutine.Start(
                    funcList,
                    op =>
                    {
                        if (op.CompletedSuccessfully)
                        {
                            TestComplete();
                        } else
                        {
                            op.MarkErrorAsHandled();
                            throw op.Error;
                        }
                    });
            });
        }

        /// <summary>
        /// Make list of asynchronous functions to be executed serially
        /// </summary>
        private IEnumerable<Func<INotifyCompleted>> MakeSerialAsyncFunctionList()
        {
            var funcList = new List<Func<INotifyCompleted>>();

            // query for customers with non-null Countries
            var custQuery = Manager.Customers.Where(c => c.Country != null);

            funcList.Add(() => custQuery.ExecuteAsync());

            // Make a change to them and save them
            funcList.Add(() =>
            {
                UpdateCachedCustomers();
                return Manager.SaveChangesAsync();
            });

            // Clear, re-load
            funcList.Add(() =>
            {
                Manager.Clear();
                return custQuery.ExecuteAsync();
            });

            // Verify change then restore customers to original state
            funcList.Add(() =>
            {
                RestoreCachedCustomer();
                return Manager.SaveChangesAsync();
            });

            return funcList;
        }

        [TestMethod, Asynchronous, Timeout(20000)]
        [Tag("Integration"), Tag("Coroutine")]
        public void can_get_many_types_with_parallel_fnc_list()
        {
            DoItAsync(() => {

                Manager.GetEntityGroups()
                    .ShouldHaveCountOf(0, "Manager is not empty.");

                var funcList = new List<Func<INotifyCompleted>>
                               {
                                   () => Manager.Products.ExecuteAsync(),
                                   () => Manager.Employees.ExecuteAsync(),
                                   () => Manager.Customers.ExecuteAsync(),
                               };

                Coroutine.StartParallel (
                    funcList,
                    op =>
                    {
                        if (op.CompletedSuccessfully)
                        {
                            Manager.GetEntityGroups()
                                .Select(g => g.EntityType)
                                .ShouldContainAllOf(
                                   new[] { 
                                       typeof(Product), 
                                       typeof(Employee), 
                                       typeof(Customer)},
                                    "Manager doesn't have expected entity types.");

                            TestComplete();
                        }
                        else
                        {
                            op.MarkErrorAsHandled();
                            throw op.Error;
                        }
                    });
            });
        }

        #region Helpers

        private const string _testSuffix = "###";

        private void UpdateCachedCustomers()
        {
            Manager
                .FindEntities<Customer>(EntityState.Unchanged)
                .ForEach(c => c.Address = c.Address + _testSuffix);
        }

        private void RestoreCachedCustomer()
        {
            Manager
                .FindEntities<Customer>(EntityState.Unchanged)
                .ForEach(c =>
                         {
                             var addr = c.Address;
                             // Fail if any address is missing the test suffix
                             Assert.IsTrue(addr.Contains(_testSuffix),
                                           c.CompanyName + " address is missing text suffix.");
                             c.Address = c.Address.Substring(0, c.Address.IndexOf(_testSuffix));
                         });
        }

        #endregion

    }
}