using System.Linq;
using Microsoft.Silverlight.Testing;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using SoftwareApproach.TestingExtensions;

using IdeaBlade.EntityModel;
using IdeaBlade.TestFramework;
using SimpleNorthwindModel;

// ReSharper disable InconsistentNaming

namespace Querying.EasyQuery {
  [TestClass]
  public class When_trying_simple_asynchronous_queries : DevForceTest {

    /// <summary>Executed before every test method.</summary>
    protected override void Context() {
      // Create manager that connects to the server automatically
      Manager = new NorthwindManager();
      PinManagerToSyncContextThread(Manager);
    }

    private NorthwindManager Manager { get; set; }
    private const int NancyDavolioId = 1;

    [TestMethod, Asynchronous, Timeout(20000)]
    [Tag("Integration"), Tag("EasyQuery")]
    public void then_get_customers_with_Model_EntityManager() {

      DoItAsync(() =>

        Manager
          .Customers // "Get all customers" query
          .ExecuteAsync()
          .TestTheResults(
            custs => custs.ShouldNotBeEmpty("Should have customers"),
            "Customers query")

       );
    }

    [TestMethod, Asynchronous, Timeout(20000)]
    [Tag("Integration"), Tag("EasyQuery")]
    public void then_get_customers_with_DevForce_EntityManager()
    {
        DoItAsync(() => 
        {
        // DevForce EntityManager knows nothing about the model
        var dfManager = new EntityManager();

        dfManager
            .GetQuery<Customer>() // "Get all customers" query
            .ExecuteAsync()
            .TestTheResults(
                custs => custs.ShouldNotBeEmpty("Should have customers"),
                "Customers query");

        });
    }

    [TestMethod, Asynchronous, Timeout(20000)]
    [Tag("Integration"), Tag("EasyQuery")]
    public void then_get_customers_verbose() {

      DoItAsync(() =>

        Manager
          .Customers
          .ExecuteAsync()
          // Same query but with "TestTheResults" inline
                  .OnCancelled(() => Assert.Fail("Unexpected cancellation of Customers query" ))
                  .OnError(err => Assert.Fail( "Customers query failed with exception: " + err))
                  .OnSuccess(custs => custs.ShouldNotBeEmpty("Should have customers"))
                  .OnCompleted( args =>  WaitFor.TestComplete())
      );
    }

    [TestMethod, Asynchronous, Timeout(20000)]
    [Tag("Integration"), Tag("EasyQuery")]
    public void then_get_Nancy_Davolio_with_fluent_syntax() {

      DoItAsync(() => {

        // Fluent syntax for query for employees with ID == to Nancy's ID
        var nancyQuery = Manager.Employees
          .Where(e => e.EmployeeID == NancyDavolioId);

        // Only want one (the first)
        // Turn into "Scalar" query and ask for First (or null if none)
        // "queryOp" is an EntityQueryOperation<Employee>
        var queryOp =
          nancyQuery
            .AsScalarAsync()
            .FirstOrDefault()

            // TestTheResult is a DevForce Operation Extension for testing
          // Put your tests inside its parameters

            .TestTheResult(nancy => {
              nancy.ShouldNotBeNullOrNullEntity("Can't find Nancy?");
              nancy.LastName.ShouldEqual("Davolio");

            });

      });
    }


    [TestMethod, Asynchronous, Timeout(20000)]
    [Tag("Integration"), Tag("EasyQuery")]
    public void then_get_Nancy_Davolio_with_query_syntax() {

      DoItAsync(() => {

        // "Query Expression" or "Comprehension" syntax
        // Query for employees with ID == to Nancy's ID
        var nancyQuery =
                from emp in Manager.Employees
                where emp.EmployeeID == NancyDavolioId
                select emp;

        // Test the "Scalar" query that asks for First (or null if none)
        nancyQuery
          .AsScalarAsync()
          .FirstOrDefault()

          // Verbose instead of TestTheResult
          .OnCancelled(() => Assert.Fail("Unexpected cancellation"))
          .OnError(err => Assert.Fail("Query failed with exception: " + err))
          .OnSuccess(nancy => {
            nancy.EntityAspect.IsNullEntity.ShouldBeFalse("Can't find Nancy?");
            nancy.LastName.ShouldEqual("Davolio");
          })
          .OnCompleted(args => TestComplete());

      });
    }


    [TestMethod, Asynchronous, Timeout(20000)]
    [Tag("Integration"), Tag("EasyQuery")]
    public void then_get_Nancy_Davolio_with_FirstOrNullEntity() {

      DoItAsync(() => {

        // Fluent syntax
        var nancyQuery = Manager.Employees
          .Where(e => e.EmployeeID == NancyDavolioId);

        nancyQuery
          .AsScalarAsync()
          .FirstOrNullEntity()

          // Verbose instead of TestTheResult
          .OnCancelled(() => Assert.Fail("Unexpected cancellation"))
          .OnError(err => Assert.Fail("Query failed with exception: " + err))
          .OnSuccess(nancy => {
            nancy.EntityAspect.IsNullEntity.ShouldBeFalse("Can't find Nancy?");
            nancy.LastName.ShouldEqual("Davolio");
          })
          .OnCompleted(args => TestComplete());

      });

    }

    [TestMethod, Asynchronous, Timeout(20000)]
    [Tag("Integration"), Tag("EasyQuery")]
    public void then_get_Nancys_orders_using_Callback() {
      DoItAsync(() =>

        Manager
          .Orders
          .Where(o => o.EmployeeID == NancyDavolioId)
          .ExecuteAsync(op => {
            op.HasError.ShouldBeFalse("Query failed with exception: " + op.Error);
            op.Results.ShouldNotBeEmpty("No orders for Nancy?");
            TestComplete();
          })

        );

    }

    [TestMethod, Asynchronous, Timeout(20000)]
    [Tag("Integration"), Tag("EasyQuery")]
    public void then_get_Nancys_orders_via_Completed_event() {

      DoItAsync(() =>
        Manager
          .Orders
          .Where(o => o.EmployeeID == NancyDavolioId)
          .ExecuteAsync()
          .Completed += (s, args) => {
            args.HasError.ShouldBeFalse("Query failed with exception: " + args.Error);
            args.Results.ShouldNotBeEmpty("No orders for Nancy?");
            TestComplete();
          }
      );

    }

  }
}
