﻿Namespace IdeaBlade.EntityModel
  Public Class OperationOnCompletedHandler(Of TOperation As BaseOperation, TEventArgs As AsyncEventArgs)

	Public Sub New(ByVal op As TOperation)
      Operation = op
	  ClearActions()
	End Sub

	Private privateOperation As TOperation
	Protected Property Operation() As TOperation
		Get
			Return privateOperation
		End Get
		Private Set(ByVal value As TOperation)
			privateOperation = value
		End Set
	End Property

	Public Overridable Sub CompletedHandler(ByVal sender As Object, ByVal args As TEventArgs)
	  Try
		If Nothing Is args Then
		  Throw New ArgumentNullException("args")
		End If
		CompletedEventArgs = args
		DoOnCancelled()
		DoOnError()
		DoOnSuccess()
		DoOnCompleted()
	  Finally
		ClearActions()
	  End Try
	End Sub

	#Region "DoOn ..."

	Protected Overridable Sub DoOnCancelled(Optional ByVal action As Action = Nothing)
	  If CompletedEventArgs.Cancelled Then
		_onCancelledActions()
		If Nothing IsNot action Then
			action()
		End If
	  End If
	End Sub

	Protected Overridable Sub DoOnError(Optional ByVal action As Action(Of Exception) = Nothing)
	  If CompletedEventArgs.HasError Then
		Operation.MarkErrorAsHandled()
		_onErrorActions(CompletedEventArgs.Error)
		If Nothing IsNot action Then
			action(CompletedEventArgs.Error)
		End If
	  End If
	End Sub

	Protected Overridable Sub DoOnSuccess(Optional ByVal action As Action(Of TEventArgs) = Nothing)
	  If CompletedEventArgs.CompletedSuccessfully Then
		_onSuccessActions(CompletedEventArgs)
		If Nothing IsNot action Then
			action(CompletedEventArgs)
		End If
	  End If
	End Sub

	Protected Overridable Sub DoOnCompleted(Optional ByVal action As Action(Of TEventArgs) = Nothing)
	  _onCompletedActions(CompletedEventArgs)
	  If Nothing IsNot action Then
		  action(CompletedEventArgs)
	  End If
	End Sub

	#End Region

	#Region "On ..."

	Public Overridable Sub OnCancelled(ByVal action As Action)
	  If WasCompletedHandlerCalled Then
		DoOnCancelled(action)
	  Else
		If Nothing Is action Then
			Return
		End If
		_onCancelledActions = DirectCast(System.Delegate.Combine(_onCancelledActions, action), Action)
	  End If
	End Sub

	Public Overridable Sub OnCompleted(ByVal action As Action(Of TEventArgs))
	  If WasCompletedHandlerCalled Then
		DoOnCompleted(action)
	  Else
		If Nothing Is action Then
			Return
		End If
		_onCompletedActions = DirectCast(System.Delegate.Combine(_onCompletedActions, action), Action(Of TEventArgs))
	  End If
	End Sub

	Public Overridable Sub OnError(ByVal failureMessageTemplate As String)
	  Dim action = FailureMessageAction(failureMessageTemplate)
      OnError(action.ToString())
	End Sub

	''' <summary>
	''' Function taking a string to format the operation exception message and returning 
	''' an action that will take an exception, format it, and do something.
	''' </summary>
	''' <remarks>
	''' The default implementation will format the operation exception message
	''' and throw an <see cref="EntityServerException"/> with this message and the
	''' incoming exception in its inner exception.
	''' <para>
	''' <emphasis>
	''' Beware: may change the type of this exception in future as it
	''' may not always true that the EntityServer is the source.</emphasis>
	''' </para><para>
	''' You can replace to give different behavior in your environment.
	''' </para>
	''' </remarks>
	Public Shared FailureMessageAction As Func(Of String, Action(Of Exception)) = Function(fmt) Sub(ex) Throw New EntityServerException(String.Format(fmt, ex.GetType().Name & " - " & ex.Message), ex)

    Public Overridable Sub OnError(ByVal action As Action(Of Exception))
      If WasCompletedHandlerCalled Then
        DoOnError(action)
      Else
        If Nothing Is action Then
          Return
        End If
        _onErrorActions = DirectCast(System.Delegate.Combine(_onErrorActions, action), Action(Of Exception))
      End If
    End Sub

    Public Overridable Sub OnSuccess(ByVal action As Action(Of TEventArgs))
      If WasCompletedHandlerCalled Then
        DoOnSuccess(action)
      Else
        If Nothing Is action Then
          Return
        End If
        _onSuccessActions = DirectCast(System.Delegate.Combine(_onSuccessActions, action), Action(Of TEventArgs))
      End If
    End Sub

#End Region


    Protected ReadOnly Property WasCompletedHandlerCalled() As Boolean
      Get
        Return Nothing IsNot CompletedEventArgs
      End Get
    End Property
    Private privateCompletedEventArgs As TEventArgs
    Protected Property CompletedEventArgs() As TEventArgs
      Get
        Return privateCompletedEventArgs
      End Get
      Private Set(ByVal value As TEventArgs)
        privateCompletedEventArgs = value
      End Set
    End Property

    Private Sub ClearActions()
      _onCancelledActions = Sub()
                            End Sub
      _onCompletedActions = Sub()
                            End Sub
      _onErrorActions = Sub()
                        End Sub
      _onSuccessActions = Sub()
                          End Sub
    End Sub

    Private _onCancelledActions As Action
    Private _onCompletedActions As Action(Of TEventArgs)
    Private _onErrorActions As Action(Of Exception)
    Private _onSuccessActions As Action(Of TEventArgs)
  End Class
End Namespace