Imports System.ComponentModel.DataAnnotations
Imports IdeaBlade.Core
Imports IdeaBlade.EntityModel
Imports IdeaBlade.TestFramework
Imports Microsoft.Silverlight.Testing
Imports Microsoft.VisualStudio.TestTools.UnitTesting
Imports SimpleNorthwindModel
Imports SoftwareApproach.TestingExtensions

' ReSharper disable InconsistentNaming
' ReSharper disable ConvertClosureToMethodGroup
' ReSharper disable MemberCanBeMadeStatic.Local
' ReSharper disable SuggestUseVarKeywordEvident

Namespace Querying.AsyncTesting


  ''' <summary>Demonstrate async queries with real database.</summary>
  <TestClass>
  Public Class When_querying_asynchronously
	  Inherits DevForceTest

	Protected Overrides Sub Context()
	  _manager = Nothing
	End Sub

	''' <summary>Record async operation coordination object for debugging purposes.</summary>
	Private Shared Property BaseOpProbe() As BaseOperation

	#Region "Manager"

	Private ReadOnly Property Manager() As NorthwindManager
	  Get
		  If _manager IsNot Nothing Then
			  Return _manager
		  Else
			  _manager = New NorthwindManager()
			  Return _manager
		  End If
	  End Get
	End Property
	Private _manager As NorthwindManager

	#End Region

	#Region "Existence query using Count, First, and First-with-projection"

	Private Const EXISTING_CUSTOMER_NAME As String = "Simons bistro"
	Private Const NONEXISTING_CUSTOMER_NAME As String = "FooBarBaz"

	<TestMethod, Asynchronous, Timeout(10000), Tag("ExistenceQuery")>
	Public Sub then_can_use_Any_in_a_Sync_query_to_find_if_customer_exists()

		' Not an async test but we structure as if it were because
		' we're running it in an async test framework
		' Ensure manager is disconnected and only runs locally
		' so we can run synchronous queries even in Silverlight
		' Attach a test entity that will satisfy the selection criteria
		  '.Any(c => c.CompanyName == NONEXISTING_CUSTOMER_NAME) // to see test fail
      DoItAsync(Sub()
                  Dim manager = New NorthwindManager(False)
                  manager.DefaultQueryStrategy = QueryStrategy.CacheOnly
                  manager.AttachEntity(New Customer With {.CustomerID = Guid.NewGuid(), .CompanyName = EXISTING_CUSTOMER_NAME})
                  manager.Customers.Any(Function(c) c.CompanyName = EXISTING_CUSTOMER_NAME).ShouldBeTrue("Should find the EXISTING CUSTOMER")
                  TestComplete()
                End Sub)
	End Sub


	 ' There is no scalar async "Any" command; three workarounds
	' Would be useful if for existence queries such as "Does any User have this UserName?"

	''' <remarks>Without an "any" command, we use Count and test for zero.</remarks>
	<TestMethod, Asynchronous, Timeout(15000), Tag("ExistenceQuery")>
	Public Sub then_can_use_Count_to_find_if_customer_exists()

      '.Count(c => c.CustomerName == NONEXISTING_CUSTOMER_NAME) // to see test fail

      BaseOpProbe = Manager.Customers.AsScalarAsync().Count(Function(c) c.CompanyName = EXISTING_CUSTOMER_NAME).TestTheResult(Function(result) result.ShouldBeGreaterThan(0, "Count of custs should be positive"))
      DoItAsync(Function() BaseOpProbe)
	End Sub



	''' <remarks>First eliminates risk of table scan for count.</remarks>
	<TestMethod, Asynchronous, Timeout(15000), Tag("ExistenceQuery")>
	Public Sub then_can_use_First_to_find_if_customer_exists()

      '.FirstOrDefault(c => c.CustomerName == NONEXISTING_CUSTOMER_NAME) // to see test fail
      BaseOpProbe = Manager.Customers.AsScalarAsync().FirstOrDefault(Function(c) c.CompanyName = EXISTING_CUSTOMER_NAME).TestTheResult(Function(result) result.ShouldNotBeNull("First cust should not be null"))
      DoItAsync(Function() BaseOpProbe)
	End Sub

	''' <remarks>Projection minimizes payload of the response.</remarks>
	<TestMethod, Asynchronous, Timeout(15000), Tag("ExistenceQuery")>
	Public Sub then_can_use_Projection_to_find_if_customer_exists()

      '.Where(c => c.CustomerName == NONEXISTING_CUSTOMER_NAME) // to see test fail
      BaseOpProbe = Manager.Customers.Where(Function(c) c.CompanyName = EXISTING_CUSTOMER_NAME).Select(Function(cust) "x").AsScalarAsync().FirstOrDefault().TestTheResult(Function(result) result.ShouldNotBeNull("First cust should not be null"))
      DoItAsync(Function() BaseOpProbe)
    End Sub

	#End Region

#Region "Projections"
    <TestMethod(), Asynchronous(), Timeout(15000)>
    <Tag("Projection")>
    Public Sub then_can_project_Employee_into_anonymous_type()

      ' Make externally excessible by pouring into a client type
      DoItAsync(Sub()
                  Dim projectedEmp As EmployeeProjection
                  BaseOpProbe = Manager.Employees.Select(Function(emp) New With {Key emp.EmployeeID, Key emp.FirstName, Key emp.LastName}).ExecuteAsync().TestTheResults(Sub(result)
                                                                                                                                                                           result.ShouldNotBeEmpty("Should have employee projections")
                                                                                                                                                                           Dim r1 = result.First()
                                                                                                                                                                           r1.GetType().Name.ShouldContain("Anonymous", "Should be anonymous type")
                                                                                                                                                                           r1.GetType().IsPublic.ShouldBeFalse("Anonymous types are not public")
                                                                                                                                                                           Dim bindableResult = DynamicTypeConverter.Convert(r1)
                                                                                                                                                                           bindableResult.GetType().IsPublic.ShouldBeTrue("Should be public after conversion")
                                                                                                                                                                           projectedEmp = New EmployeeProjection With {.EmployeeId = r1.EmployeeID, .FirstName = r1.FirstName, .LastName = r1.LastName}
                                                                                                                                                                           String.IsNullOrWhiteSpace(projectedEmp.FirstName).ShouldBeFalse("FirstName should have real value")
                                                                                                                                                                         End Sub)
                End Sub)
    End Sub

    ''' <summary>
    ''' Client-side type to ease access to employee slice projection
    ''' </summary>
    Public Class EmployeeProjection
      Private privateEmployeeId As Integer
      Public Property EmployeeId() As Integer
        Get
          Return privateEmployeeId
        End Get
        Friend Set(ByVal value As Integer)
          privateEmployeeId = value
        End Set
      End Property
      Private privateFirstName As String
      Public Property FirstName() As String
        Get
          Return privateFirstName
        End Get
        Friend Set(ByVal value As String)
          privateFirstName = value
        End Set
      End Property
      Private privateLastName As String
      Public Property LastName() As String
        Get
          Return privateLastName
        End Get
        Friend Set(ByVal value As String)
          privateLastName = value
        End Set
      End Property
    End Class

    ''' <summary>
    ''' Projecting Employee
    ''' into a custom type, <see cref="EmployeeSlice"/>,
    '''  defined on both server and client.
    ''' </summary>
    <TestMethod(), Asynchronous(), Timeout(15000), Tag("Projection")>
    Public Sub then_can_project_Employee_into_EmployeeSlice()
      BaseOpProbe = Manager.Employees.Select(Function(emp) New EmployeeSlice With {.EmployeeID = emp.EmployeeID, .FirstName = emp.FirstName, .LastName = emp.LastName}).ExecuteAsync().TestTheResults(Sub(result)
                                                                                                                                                                                                        result.ShouldNotBeEmpty("Should have EmployeeSlice objects")
                                                                                                                                                                                                        String.IsNullOrWhiteSpace(result.First().FirstName).ShouldBeFalse("FirstName should have real value")
                                                                                                                                                                                                      End Sub)

      DoItAsync(Function() BaseOpProbe)
    End Sub


    ''' <summary>
    ''' Projecting Employee and count of its orders
    ''' into a custom type, <see cref="EmployeeSlice"/>,
    '''  defined on both server and client.
    ''' </summary>
    <TestMethod(), Asynchronous(), Timeout(15000), Tag("Projection")>
    Public Sub then_can_project_Employee_and_OrderCount_into_EmployeeSlice()
      BaseOpProbe = Manager.Employees.Select(Function(emp) New EmployeeSlice With {.EmployeeID = emp.EmployeeID, .FirstName = emp.FirstName, .LastName = emp.LastName, .OrderCount = emp.Orders.Count()}).ExecuteAsync().TestTheResults(Sub(result)
                                                                                                                                                                                                                                          result.ShouldNotBeEmpty("Should have EmployeeSlice objects")
                                                                                                                                                                                                                                          result.First().OrderCount.ShouldBeGreaterThan(0, "Should have projected count of emp orders")
                                                                                                                                                                                                                                        End Sub)
    End Sub

#End Region

#Region "Contains (IN)"

    <TestMethod(), Asynchronous(), Timeout(15000), Tag("AsyncTesting"), Tag("Integration")>
    Public Sub then_can_get_employees_asynchronously_by_id_list()

      DoItAsync(Sub()
                  Dim idArray = {1, 3, 5}
                  Dim idList = idArray.ToList()
                  Dim op = Manager.Employees.Where(Function(emp) idList.Contains(emp.EmployeeID)).ExecuteAsync()
                  AddHandler op.Completed, Sub(s, args)
                                             If (args.HasError) Then
                                               Assert.IsFalse(args.HasError, "Query exception: " & args.Error.ToString())
                                             Else
                                               Assert.IsFalse(args.HasError, "Query exception: ")
                                             End If
                                             Dim emps = args.Results
                                             emps.Count().ShouldEqual(3)
                                             TestComplete()
                                           End Sub
                End Sub)

    End Sub

    ''' <summary>
    ''' Query with 'contains', the LINQ equivalent of the SQL 'IN' operator.
    ''' </summary>
    <TestMethod(), Asynchronous(), Timeout(15000)>
    Public Sub then_can_get_customers_in_Europe()

      ' Prepare a LIST (not an ARRAY) of european countries
      ' The LINQ equivalent of IN is CONTAINS
      DoItAsync(Sub()
                  Dim countryNames = New List(Of String) From {"UK", "France"}
                  Dim query = Manager.Customers.Where(Function(c) countryNames.Contains(c.Country))
                  query.ExecuteAsync().OnError(Sub(err) Assert.Fail("Contains ('IN') query failed: " & err.ToString())).OnSuccess(Sub(euroCusts)
                                                                                                                                    euroCusts.ShouldNotBeEmpty("No euro customers")
                                                                                                                                    euroCusts.Any(Function(c) c.Country = "USA").ShouldBeFalse()
                                                                                                                                    euroCusts.Any(Function(c) c.Country = "UK").ShouldBeTrue()
                                                                                                                                  End Sub).OnCompleted(Sub() TestComplete())
                End Sub)
    End Sub

    <TestMethod(), Asynchronous(), Timeout(15000)>
    Public Sub then_fails_using_an_array_in_Contains()

      ' WRONG: don't use an *array* of european countries
      ' The LINQ equivalent of IN is CONTAINS
      DoItAsync(Sub()
                  Dim countryNamesArray() As String = {"UK", "France"}
                  Dim query = Manager.Customers.Where(Function(c) countryNamesArray.Contains(c.Country))
                  BaseOpProbe = query.ExecuteAsync().ExpectException(Of InvalidOperationException)()
                End Sub)
    End Sub
#End Region

#Region "String match case sensitivity"

    ' Technique required until DF 6.0.5
    <TestMethod(), Asynchronous(), Timeout(15000), Tag("AsyncTesting"), Tag("Integration")>
    Public Sub then_can_get_employees_with_case_insensitive_EQUALS_method_query()

      DoItAsync(Sub()
                  Const lastNameToMatch As String = "davolio"
                  Dim op = Manager.Employees.Where(Function(emp) emp.LastName.Equals(lastNameToMatch, StringComparison.CurrentCultureIgnoreCase)).AsScalarAsync().FirstOrDefault()
                  AddHandler op.Completed, Sub(s, args)
                                             If (args.HasError) Then
                                               Assert.IsFalse(args.HasError, "Query exception: " & args.Error.ToString())
                                             Else
                                               Assert.IsFalse(args.HasError, "Query exception: ")
                                             End If
                                             Dim nancy = args.Result
                                             nancy.ShouldNotBeNull("No employee matching " & lastNameToMatch)
                                             nancy.LastName.ShouldEqual("Davolio", "Last name was not 'Davolio'")
                                             nancy.LastName.ShouldNotEqual(lastNameToMatch, "Supposed to be testing case-insensitivity")
                                             TestComplete()
                                           End Sub
                End Sub)

    End Sub

    ' Implied case-insensitivity beginning with DF 6.0.5
    <TestMethod(), Asynchronous(), Timeout(15000), Tag("AsyncTesting"), Tag("Integration")>
    Public Sub then_can_get_employees_with_case_insensitive_query()

      DoItAsync(Sub()
                  Const lastNameToMatch As String = "davolio"
                  Dim op = Manager.Employees.Where(Function(emp) emp.LastName.Equals(lastNameToMatch)).AsScalarAsync().FirstOrDefault()
                  AddHandler op.Completed, Sub(s, args)
                                             If (args.HasError) Then
                                               Assert.IsFalse(args.HasError, "Query exception: " & args.Error.ToString())
                                             Else
                                               Assert.IsFalse(args.HasError, "Query exception: ")
                                             End If
                                             Dim nancy = args.Result
                                             If (nancy Is Nothing) Then
                                               Assert.Fail("Nancy is nothing")
                                             Else
                                               nancy.ShouldNotBeNull("No employee matching " & lastNameToMatch)
                                               nancy.LastName.ShouldEqual("Davolio", "Last name was not 'Davolio'")
                                               nancy.LastName.ShouldNotEqual(lastNameToMatch, "Supposed to be testing case-insensitivity")
                                             End If
                                             TestComplete()
                                           End Sub
                End Sub)
    End Sub

#End Region

#Region "Exclude Entities Returned to the Client"

    ''' <summary>
    ''' <see cref="EntityManager.Queried"/> event handler looks for 
    ''' non-Brazilian orders returned after a data source fetch
    ''' and removes them from the cache before the client can see or use them.
    ''' Does so without disturbing the query cache which means that
    ''' previously cached queries will continue to return entities from
    ''' the entity cache w/o going to the server (unless told to do so).
    ''' </summary>
    Private Sub BrazilOnlyQueriedHandler(ByVal sender As Object, ByVal e As EntityQueriedEventArgs)
      ' Remove retrieved non-Brazilian orders from entity cache
      ' without disturbing the query cache.
      Manager.RemoveEntities(e.ChangedEntities.OfType(Of Order)().Where(Function(o) o.ShipCountry <> "Brazil"), False) ' don't clear query cache
    End Sub

    ''' <summary>
    ''' Placeholder to preserve structure of all tests;
    ''' you would not add this handler in real world.
    ''' </summary>
    Private Sub DoNothingQueriedHandler(ByVal sender As Object, ByVal e As EntityQueriedEventArgs)
    End Sub

    <TestMethod(), Asynchronous(), Timeout(15000)>
    Public Sub then_have_queried_Orders_from_outside_Brazil_when_no_QueriedHandler()

      DoItAsync(Sub()
                  AddHandler Manager.Queried, AddressOf DoNothingQueriedHandler
                  Manager.Orders.ExecuteAsync().OnError(Sub(ex) Assert.Fail("Order query failed with: " & ex.ToString())).OnSuccess(Sub(args)
                                                                                                                                      Dim ordersInCache = Manager.FindEntities(GetType(Order), EntityStates.AllCachedStates).OfType(Of Order)()
                                                                                                                                      Dim nonBrazilOrders = ordersInCache.Where(Function(o) o.ShipCountry <> "Brazil")
                                                                                                                                      nonBrazilOrders.ShouldNotBeEmpty("Should have non-Brazilian orders")
                                                                                                                                    End Sub).OnCompleted(Sub() TestComplete())
                End Sub)
    End Sub

    <TestMethod(), Asynchronous(), Timeout(15000)>
    Public Sub then_no_queried_Orders_from_outside_Brazil_with_QueriedHandler()

      DoItAsync(Sub()
                  AddHandler Manager.Queried, AddressOf BrazilOnlyQueriedHandler
                  Manager.Orders.ExecuteAsync().OnError(Sub(ex) Assert.Fail("Order query failed with: " & ex.ToString())).OnSuccess(Sub(args)
                                                                                                                                      Dim ordersInCache = Manager.FindEntities(GetType(Order), EntityStates.AllCachedStates).OfType(Of Order)()
                                                                                                                                      Dim nonBrazilOrders = ordersInCache.Where(Function(o) o.ShipCountry <> "Brazil")
                                                                                                                                      nonBrazilOrders.ShouldBeEmpty("Should only have Brazilian orders")
                                                                                                                                    End Sub).OnCompleted(Sub() TestComplete())
                End Sub)
    End Sub

    <TestMethod(), Asynchronous(), Timeout(15000)>
    Public Sub then_have_included_Orders_from_outside_Brazil_when_no_QueriedHandler()

      DoItAsync(Sub()
                  AddHandler Manager.Queried, AddressOf DoNothingQueriedHandler
                  Manager.Employees.Include(Employee.EntityPropertyNames.Orders).ExecuteAsync().OnError(Sub(ex) Assert.Fail("Order query failed with: " & ex.ToString())).OnSuccess(Sub(args)
                                                                                                                                                                                      Dim ordersInCache = Manager.FindEntities(GetType(Order), EntityStates.AllCachedStates).OfType(Of Order)()
                                                                                                                                                                                      Dim nonBrazilOrders = ordersInCache.Where(Function(o) o.ShipCountry <> "Brazil")
                                                                                                                                                                                      nonBrazilOrders.ShouldNotBeEmpty("Should have non-Brazilian orders")
                                                                                                                                                                                    End Sub).OnCompleted(Sub() TestComplete())
                End Sub)
    End Sub

    <TestMethod(), Asynchronous(), Timeout(20000)>
    Public Sub then_entity_removing_queried_handler_preserves_query_caching()

      ' Prove that QueryHandler did remove all incoming entities
      ' Assert failure if this manager ever fetches from remote store again
      ' Repeat query - which should be remembered and execute CacheOnly
      ' if no failures, the repeated query must have succeeded CacheOnly
      DoItAsync(Sub()
                  AddHandler Manager.Queried, AddressOf EntityRemovingQueriedHandler
                  Dim query = Manager.Orders
                  query.ExecuteAsync().OnError(Sub(ex)
                                                 Assert.Fail("Order query failed with: " & ex.ToString())
                                               End Sub).OnSuccess(Sub(result)
                                                                    Dim cachedEntities = Manager.FindEntities(EntityStates.AllCachedStates).OfType(Of Object)()
                                                                    cachedEntities.ShouldBeEmpty("Queried event handler should have removed everything")
                                                                    AddHandler Manager.Fetching, Sub() Assert.Fail("Should not re-fetch")
                                                                    query.ExecuteAsync().OnError(Sub(ex) Assert.Fail("Re-query failed: " & ex.ToString())).OnCompleted(Sub() TestComplete())
                                                                  End Sub)
                End Sub)
    End Sub


    Private Sub EntityRemovingQueriedHandler(ByVal sender As Object, ByVal e As EntityQueriedEventArgs)
      ' Removes ALL entities that would be affected by this query
      ' This flag is CRITICAL - test fails if 'true' 
      Manager.RemoveEntities(e.ChangedEntities, False) ' don't clear query cache
    End Sub
#End Region

#Region "Miscellaneous"
    ' ReSharper disable ConvertToLambdaExpression

    <TestMethod(), Asynchronous(), Timeout(20000)>
    Public Sub then_async_nav_to_customer_of_order_does_not_change_order()


#If Not SILVERLIGHT Then
	  DoItAsync(Function() {Manager.UseAsyncNavigation = true; Manager ' for this test
#Else
      DoItAsync(Sub()
#End If
                  Manager.Orders.AsScalarAsync().First().OnError(Sub(err) Assert.Fail("Order query failed: " & err.ToString())).OnSuccess(Sub(order)
                                                                                                                                            order.CustomerID.ShouldNotBeNull("order.CustomerID should not be null after fetch")
                                                                                                                                            Dim cust = order.Customer
                                                                                                                                            cust.EntityAspect.IsPendingEntity.ShouldBeTrue("order.Customer is not a pending entity.")
                                                                                                                                            order.CustomerID.ShouldNotBeNull("CustomerID should not be null after customer navigation")
                                                                                                                                            order.ShouldNotBeChanged("; pending customer modified the order")
                                                                                                                                            AddHandler cust.EntityAspect.PendingEntityResolved, Sub(s, args)
                                                                                                                                                                                                  Dim resolvedcust As Customer = CType(args.ResolvedEntity, Customer)
                                                                                                                                                                                                  resolvedcust.ShouldNotBeSameAs(cust, "Resolved cust entity should be different than cust")
                                                                                                                                                                                                  order.CustomerID.ShouldNotBeNull("CustomerID should not be null after resolution")
                                                                                                                                                                                                  order.ShouldNotBeChanged("; pending customer modified the order")
                                                                                                                                                                                                  TestComplete()
                                                                                                                                                                                                End Sub
                                                                                                                                          End Sub)
                End Sub)
    End Sub

    <TestMethod(), Asynchronous(), Timeout(20000), Tag("Pending")>
    Public Sub then_no_async_nav_to_customer_of_order_within_NORMAL_query()
#If Not SILVERLIGHT Then
	  DoItAsync(Function() {Manager.UseAsyncNavigation = true; Manager
	  DoItAsync(Function() { Manager
#End If
      Manager.Orders.AsScalarAsync().First().OnError(Sub(err) Assert.Fail("Order query failed: " & err.ToString())).OnSuccess(Sub(order)
                                                                                                                                Dim custId = order.CustomerID
                                                                                                                                custId.ShouldNotBeNull("order.CustomerID should not be null after fetch")
                                                                                                                                Manager.Orders.Where(Function(o) o.Customer.CustomerID = custId).With(QueryStrategy.Normal).ExecuteAsync().OnError(Sub(err) Assert.Fail("Orders-of-CustId failed: " & err.ToString())).OnSuccess(Sub(results)
                                                                                                                                                                                                                                                                                                                                   results.ShouldNotBeEmpty("No orders-of-CustId")
                                                                                                                                                                                                                                                                                                                                   Dim cust = If(results.First().Customer, Manager.GetNullEntity(Of Customer)())
                                                                                                                                                                                                                                                                                                                                   cust.EntityAspect.IsNullOrPendingEntity.ShouldBeFalse("cust should be a real entity, returned by query inversion.")
                                                                                                                                                                                                                                                                                                                                   TestComplete()
                                                                                                                                                                                                                                                                                                                                 End Sub)
                                                                                                                              End Sub)
    End Sub

    <TestMethod(), Asynchronous(), Timeout(20000), Tag("Pending")>
    Public Sub then_no_async_nav_to_customer_of_order_within_CACHEONLY_query()
      DoItAsync(Sub()
                  Dim firstOrder As Order
                  Manager.Orders.AsScalarAsync().First().OnError(Sub(err) Assert.Fail("Order query failed: " & err.ToString())).OnSuccess(Sub(order)
                                                                                                                                            firstOrder = order
                                                                                                                                            Dim custId = firstOrder.CustomerID
                                                                                                                                            custId.ShouldNotBeNull("order.CustomerID should not be null after fetch")
                                                                                                                                            Manager.Orders.Where(Function(o) o.Customer.CustomerID = custId).With(QueryStrategy.CacheOnly).ExecuteAsync().OnError(Sub(err) Assert.Fail("Orders-of-Customer-with-CustId failed: " & err.ToString())).OnSuccess(Sub(results)
                                                                                                                                                                                                                                                                                                                                                                results.ShouldBeEmpty("Should have no Orders-of-Customer-with-CustId")
                                                                                                                                                                                                                                                                                                                                                                Manager.Orders.Where(Function(o) o.CustomerID = custId).With(QueryStrategy.CacheOnly).ExecuteAsync().OnError(Sub(err) Assert.Fail("Orders-with-CustId failed: " & err.ToString())).OnSuccess(Sub(results2)
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                               results2.Count().ShouldEqual(1, "Only one order should be in cache.")
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                               results2.ShouldContain(firstOrder, "'firstOrder' should be the only order")
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                               TestComplete()
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             End Sub)
                                                                                                                                                                                                                                                                                                                                                              End Sub)
                                                                                                                                          End Sub)
                End Sub)
    End Sub

    ' ReSharper restore ConvertToLambdaExpression
#End Region

#Region "Poco"

    <TestMethod(), Asynchronous(), Timeout(15000)>
    Public Sub then_can_query_Poco_in_cache()
      DoItAsync(Sub()
                  Dim manager = New EntityManager(False)
                  Dim query = New EntityQuery(Of DummyEntity)().With(QueryStrategy.CacheOnly)
                  AddHandler manager.ExecuteQueryAsync(query).Completed, (Sub(s, args)
                                                                            If (args.HasError) Then
                                                                              Assert.IsFalse(args.HasError, "Query failed with " & args.Error.ToString())
                                                                            Else
                                                                              Assert.IsFalse(args.HasError, "Query failed with ")
                                                                            End If
                                                                            TestComplete()
                                                                          End Sub)
                End Sub)
    End Sub

    Public Class DummyEntity
      <Key()>
      Public Property Id() As Integer
    End Class

#End Region

  End Class

End Namespace