Imports System.Collections.ObjectModel
Imports System.Collections.Specialized
Imports Microsoft.Silverlight.Testing
Imports Microsoft.VisualStudio.TestTools.UnitTesting
Imports SoftwareApproach.TestingExtensions

Imports IdeaBlade.Core
Imports IdeaBlade.EntityModel
Imports SimpleNorthwindModel
Imports IdeaBlade.TestFramework

' ReSharper disable InconsistentNaming
' ReSharper disable SuggestUseVarKeywordEvident
' ReSharper disable PartialTypeWithSinglePart
' ReSharper disable RedundantAssignment

'INSTANT VB TODO TASK: There is no equivalent to #pragma directives in VB:
'#pragma warning disable 168 ' unused variable


Namespace Querying.CacheOnly

  <TestClass>
  Partial Public Class When_querying_cache_only
	  Inherits CacheOnlyContext

	''' <summary>Illustrate variety of ways to execute a query synchronously</summary>
	<TestMethod>
	Public Sub then_can_get_all_employees()

	  ' Define a query to get all employees
	  Dim query = Manager.Employees

	  ' Execute it with "ToList()"
	  Dim empsList As List(Of Employee) = query.ToList()
	  empsList.ShouldNotBeEmpty("Should have employees")

	  ' Same collection query run with variety of alternative immediate execution commands.

	  Dim empsEnumerable As IEnumerable(Of Employee) = query.Execute()
	  empsEnumerable.ShouldNotBeEmpty("Should have employees after Execute")

	  Dim empsArray() As Employee = query.ToArray()
	  empsArray.ShouldNotBeEmpty("Should have employees after ToArray")

      Dim empDictionary As IDictionary(Of Integer, Employee) = query.ToDictionary(Function(emp) emp.EmployeeID)
      empDictionary.Any().ShouldBeTrue("Should have employees after ToDictionary")

	End Sub

	<TestMethod>
	Public Sub then_can_get_customers_in_Europe()

	  ' Prepare a LIST (not an ARRAY) of european countries
	  ' DevForce deliberately throws an informative exception if you try an array.
	  Dim countryNames = New List(Of String) From {"UK", "France"}

	  ' The LINQ equivalent of IN is CONTAINS
	  Dim query = Manager.Customers.Where(Function(c) countryNames.Contains(c.Country))

	  Dim euroCusts = query.ToList()

	  euroCusts.ShouldNotBeEmpty("No euro customers")
      euroCusts.Any(Function(c) c.Country = "USA").ShouldBeFalse()
      euroCusts.Any(Function(c) c.Country = "UK").ShouldBeTrue()

	End Sub

	''' <summary>Illustrate variety of ways to execute a query synchronously</summary>
	<TestMethod>
	Public Sub then_can_get_all_companies_in_a_particular_country()

	  ' Query for customers in the USA - fluent syntax
	  Dim usaQuery = Manager.Customers.Where(Function(c) c.Country = "USA")

	  ' Query for customers in the UK - query syntax
	  Dim ukQuery = From uk In Manager.Customers
	                Where uk.Country = "UK"
	                Select uk

	  ' Execute each query and confirm they return something
	  Dim usaCusts = usaQuery.ToList()
	  usaCusts.ShouldNotBeEmpty("No customers in the US")

	  Dim ukCusts = ukQuery.ToList()
	  ukCusts.ShouldNotBeEmpty("No customers in the UK")

	End Sub

	<TestMethod>
	Public Sub then_can_find_employee_in_cache_by_entity_key()

	  ' Find by EntityKey, no matter what its EntityState
	  Dim emp As Employee = CType(Manager.FindEntity(New EntityKey(GetType(Employee), TestEmployeeId)), Employee)

	  emp.ShouldNotBeNull("Test employee not found")
	  emp.EmployeeID.ShouldEqual(TestEmployeeId)

	End Sub

	<TestMethod>
	Public Sub then_employee_not_found_in_cache_is_null()

	  ' When not found returns null
	  Dim emp = CType(Manager.FindEntity(New EntityKey(GetType(Employee), Integer.MaxValue)), Employee)

	  emp.ShouldBeNull()
	End Sub

	<TestMethod>
	Public Sub then_can_query_for_employee_in_cache_by_id()

	  ' Query by id of entity known to be in cache 
	  Dim query = Manager.Employees.Where(Function(emp) emp.EmployeeID = TestEmployeeId).With(QueryStrategy.CacheOnly)

	  Dim employee = query.SingleOrDefault()

	  employee.ShouldNotBeNullOrNullEntity("Test employee not found")
	  employee.EmployeeID.ShouldEqual(TestEmployeeId)

	End Sub

	<TestMethod>
	Public Sub then_an_employee_not_found_in_cache_by_id_query_is_nullentity()

	  ' Query by id we know is NOT in cache or in the database 
	  Dim query = Manager.Employees.Where(Function(emp) emp.EmployeeID = Integer.MaxValue).With(QueryStrategy.CacheOnly)

	  Dim employee = query.SingleOrDefault()

	  employee.ShouldBeNullOrNullEntity("employee with 'impossible' id was found")
	End Sub

	<TestMethod>
	Public Sub then_can_create_query_without_a_manager()

	  ' Create a query to get all Employees without referring to a manager
	  Dim query As New EntityQuery(Of Employee)()

	  ' Prove it works when give a manager
	  query.EntityManager.ShouldBeNull("query shouldn't have a manager")
	  Dim emps = Manager.ExecuteQuery(query)
	  emps.ShouldNotBeEmpty("Should have employees")

	End Sub

	<TestMethod>
	Public Sub then_can_reuse_query_with_different_managers()

	  ' Create a query to get all Employees without referring to a manager
	  Dim query As New EntityQuery(Of Employee)()

	  ' Execute with first manager
	  Dim emps = Manager.ExecuteQuery(query)
	  emps.ShouldNotBeEmpty("Should have employees")


	  ' Execute with second manager (using "With" verb)
	  Dim manager2 = PrepareNewOfflineManager()
	  Dim emps2 = query.With(manager2)
	  emps2.ShouldNotBeEmpty("Should have employees")

	  ' Managers are different
	  Manager.ShouldNotBeSameAs(manager2)

	  ' Employee instances are different
	  ' because the managers and caches are different
	  emps.ShouldNotContainAnyOf(emps2, "No Employee in second manager should be among first manager's employees")

	  ' But the employess are the same
	  ' from the perspective of their identities
	  emps.Select(Function(e) e.EmployeeID).ShouldContainAllOf(emps.Select(Function(e2) e2.EmployeeID), "Employee IDs in second manager should be among first manager's employee IDs")
	End Sub


	<TestMethod>
	Public Sub then_can_import_remove_and_reimport_entities()
		Dim cust = Manager.Customers.Where(Function(c) c.Orders.Any()).First()
		Dim orders = cust.Orders
		Dim q = cust.ToQuery()
		Dim em2 = New NorthwindManager(False) ' offline

		em2.ImportEntities(orders.ToList().Cast(Of Object)().Concat(cust), MergeStrategy.OverwriteChanges)
		Dim sameCust = q.With(em2).First()
		Assert.IsTrue(sameCust.EntityAspect.EntityState.IsUnchanged())

		em2.RemoveEntities(GetType(Order), EntityState.AllButDetached)
		em2.RemoveEntities(GetType(Customer), EntityState.AllButDetached)
		Assert.IsTrue(sameCust.EntityAspect.EntityState.IsDetached())

		em2.ImportEntities(orders.ToList().Cast(Of Object)().Concat(cust), MergeStrategy.OverwriteChanges)
		Dim cust2 = q.With(em2).First()
		Assert.IsTrue(cust2.EntityAspect.EntityState.IsUnchanged())
	End Sub


  End Class

  ''' <summary>
  ''' Demonstrates a custom filter navigation that returns
  ''' a live collection maintained by the EntityManager
  ''' </summary>
  <TestClass>
  Public Class WhenUsingFilteredNavigation
	  Inherits CacheOnlyContext
	  #Region "Test Initialization"

	  Protected Overrides Sub Context()
		  _cyOrdersChange = Nothing
		  CurrentYearOrders = TestEmployee.CurrentYearOrders
      AddHandler CurrentYearOrders.CollectionChanged, Sub(s, e)
                                                        _cyOrdersChange = e.Action
                                                      End Sub
	  End Sub

	  Private _cyOrdersChange? As NotifyCollectionChangedAction
	  Private Property CurrentYearOrders() As ObservableCollection(Of Order)

	  #End Region

	  ' The way it works for a generated navigation property
	  <TestMethod, Tag("Model"), Tag("CurrentYearOrders")>
	  Public Sub OrdersShouldChangeWhenOrderAttachedToManager()
		  Dim orders = TestEmployee.Orders
		  Dim ordersChange? As NotifyCollectionChangedAction = Nothing
      AddHandler orders.CollectionChanged, Sub(_sender, e) ordersChange = e.Action

		  Dim attachedOrder = New Order With {.OrderID = 100, .EmployeeID = TestEmployee.EmployeeID}

		  Manager.AttachEntity(attachedOrder)

		  Assert.IsTrue(orders.Contains(attachedOrder), "attachedOrder not among Orders")

		  Assert.IsTrue(ordersChange.HasValue, "orders did not raise change event when attaching order")
		  Assert.AreEqual(NotifyCollectionChangedAction.Add, ordersChange.Value, "orders raised unexpected change action when attaching order")

	  End Sub

	  <TestMethod, Tag("Model"), Tag("CurrentYearOrders")>
	  Public Sub ShouldHaveCurrentYearOrders()
		  Dim Orders = TestEmployee.Orders

		  Assert.AreNotEqual(0, CurrentYearOrders.Count, "CurrentYearOrders is empty")

		  Assert.IsTrue(CurrentYearOrders.Count < Orders.Count, "CurrentYearOrders count >= Orders")

		  Assert.IsFalse(CurrentYearOrders.Any(Function(c) c.OrderDate Is Nothing OrElse c.OrderDate.Value.Year <> 1998), "Some CurrentYearOrders were not ordered in 1998")
	  End Sub

	  <TestMethod, Tag("Model"), Tag("CurrentYearOrders")>
	  Public Sub ShouldFindCyOrderAttachedToManager()
		  Dim newCyOrder = New Order With {.OrderID = 100, .EmployeeID = TestEmployeeId, .OrderDate = New Date(1998, 5, 1)}

		  Manager.AttachEntity(newCyOrder)

		  Assert.IsTrue(CurrentYearOrders.Contains(newCyOrder), "newCyOrder attached to manager not among CurrentYearOrders")

		  AssertCurrentYearOrdersChanged(NotifyCollectionChangedAction.Add, "order attached to manager")

	  End Sub


	  <TestMethod, Tag("Model"), Tag("CurrentYearOrders")>
	  Public Sub ShouldFindCYOrderAddedToCurrentYearOrders()
		  Dim newCyOrder = New Order With {.OrderID = 100, .EmployeeID = TestEmployeeId, .OrderDate = New Date(1998, 4, 1)}

		  Manager.AttachEntity(newCyOrder)

		  CurrentYearOrders.Add(newCyOrder)

		  Assert.IsTrue(TestEmployee.CurrentYearOrders.Contains(newCyOrder), "newCyOrder added to CurrentYearOrders not among subsequent CurrentYearOrders")

		  AssertCurrentYearOrdersChanged(NotifyCollectionChangedAction.Add, "order added to CurrentYearOrders")

	  End Sub

	  '[Bug("1661")] We decided not to fix this
	  <TestMethod, Tag("Model"), Tag("CurrentYearOrders")>
	  Public Sub WillFindOrderAddedToCurrentYearOrdersIfNotCY()
		  Dim notCyOrder = New Order With {.OrderID = 100, .EmployeeID = TestEmployee.EmployeeID + 1, .OrderDate = New Date(1996, 1, 1)}
		  Manager.AttachEntity(notCyOrder)

		  Assert.IsFalse(CurrentYearOrders.Contains(notCyOrder), "Found notCyOrder among CurrentYearOrders before adding to CYOrders")

		  CurrentYearOrders.Add(notCyOrder)

		  'Assert.IsFalse(CurrentYearOrders.Contains(notCyOrder),// if we fixed it
		  Assert.IsTrue(CurrentYearOrders.Contains(notCyOrder), "Found notCyOrder among CurrentYearOrders after adding to CYOrders") ' because we wont


		  ' if we fixed it
		  'AssertCurrentYearOrdersDidNotChange("1996 order for different employee added.");
	  End Sub

	  ' Simplified version submitted as defect
	  '[Bug("1661")] We decided not to fix this
	  <TestMethod, Tag("Model"), Tag("CurrentYearOrders")>
	  Public Sub WillFindOrderAddedToCurrentYearOrdersIfNotCY2()
		  Dim currentYearOrders = TestEmployee.CurrentYearOrders

		  Dim notCyOrder = New Order With {.OrderID = 100, .EmployeeID = TestEmployee.EmployeeID + 1, .OrderDate = New Date(1996, 1, 1)}

		  Manager.AttachEntity(notCyOrder)

		  Assert.IsFalse(currentYearOrders.Contains(notCyOrder), "Found notCyOrder among CurrentYearOrders before adding to CYOrders")

		  currentYearOrders.Add(notCyOrder)

		  'Assert.IsFalse(currentYearOrders.Contains(notCyOrder), // if we fixed it
		  Assert.IsTrue(currentYearOrders.Contains(notCyOrder), "Found notCyOrder among CurrentYearOrders after adding to CYOrders") ' because we wont

	  End Sub

	  <TestMethod, Tag("Model"), Tag("CurrentYearOrders")>
	  Public Sub ShouldFindOrderChangedTo1998AmongCurrentYearOrders()
      Dim modOrder = TestEmployee.Orders.First(Function(o) o.OrderDate Is Nothing)

      Assert.IsFalse(CurrentYearOrders.Contains(modOrder), "modOrder has a null OrderDate but is among CurrentYearOrders")

      modOrder.OrderDate = New Date(1998, 10, 4)

      Assert.IsTrue(TestEmployee.CurrentYearOrders.Contains(modOrder), "modOrder is not among subsequent CurrentYearOrders after OrderDate set to 1998")

      AssertCurrentYearOrdersChanged(NotifyCollectionChangedAction.Add, "order with null OrderDate set to 1998")

	  End Sub

    <TestMethod(), Tag("Model"), Tag("CurrentYearOrders")>
    Public Sub ShouldNotFindOrderChangedTo1996AmongCurrentYearOrders()
      Dim modOrder = TestEmployee.Orders.First()

      modOrder.OrderDate = New Date(1996, 10, 4)

      Assert.IsFalse(TestEmployee.CurrentYearOrders.Contains(modOrder), "modOrder is among subsequent CurrentYearOrders after OrderDate set to 1996")

      ' Line below needed for VB; not necessary in C#
      CurrentYearOrders.First().OrderDate = New Date(1996, 10, 4)

      AssertCurrentYearOrdersChanged(NotifyCollectionChangedAction.Remove, "existing CY Order's OrderDate set to 1996")
    End Sub

#Region "Private asserts"

    Private Sub AssertCurrentYearOrdersChanged(ByVal expectedChange As NotifyCollectionChangedAction, ByVal whenMessage As String)
      If _cyOrdersChange.HasValue Then
        If expectedChange = _cyOrdersChange Then
          Return
        End If
        Assert.Fail(String.Format("CurrentYearOrders change action was {0} instead of {1} when {2}.", _cyOrdersChange, expectedChange, whenMessage))
      Else
        Assert.Fail("CurrentYearOrders did not raise change event when " & whenMessage)
      End If
    End Sub

    Private Sub AssertCurrentYearOrdersDidNotChange(ByVal whenMessage As String)
      If Not _cyOrdersChange.HasValue Then
        Return
      End If

      Assert.Fail(String.Format("CurrentYearOrders unexpectedly changed when {1}; action was {0}.", _cyOrdersChange, whenMessage))
    End Sub

#End Region

  End Class

  <TestClass>
  Public Class WhenModifyingEntities
	  Inherits CacheOnlyContext

#If SILVERLIGHT Then
		' INotifyDataErrorInfo only defined in Silverlight
		<TestMethod, Tag("Model"), Tag("Mods")>
		Public Sub ShouldNotHaveErrorsAfterRejectChanges()
			Dim indei = CType(TestCustomer, System.ComponentModel.INotifyDataErrorInfo)
			Assert.IsFalse(indei.HasErrors, "Should not have errors at the start")

			TestCustomer.CompanyName = Nothing
			Assert.IsTrue(indei.HasErrors, "Should have error after nulling name")

			TestCustomer.EntityAspect.RejectChanges()
			TestCustomer.EntityAspect.ValidationErrors.Clear() ' ToDo: DevForce should do this?
			Assert.IsFalse(indei.HasErrors, "Should not have errors after reject changes")
		End Sub

		<TestMethod, Tag("Model"), Tag("Mods"), Bug("#1663", Fixed := True)>
		Public Sub ShouldRaiseEventWhenClearValidationErrors()

			' Arrange (w/ some pre-lim asserts
			Dim indei = CType(TestCustomer, System.ComponentModel.INotifyDataErrorInfo)
			Assert.IsFalse(indei.HasErrors, "Should not have errors at the start")

			TestCustomer.CompanyName = Nothing
			Assert.IsTrue(indei.HasErrors, "Should have error after nulling name")

			Dim notified = False
			AddHandler indei.ErrorsChanged, Sub(s,e) notified = True

			' Act
			TestCustomer.EntityAspect.ValidationErrors.Clear()

			' Assert
			Assert.IsFalse(indei.HasErrors, "Should not have errors after clearing them")
			Assert.IsTrue(notified, "ErrorsChanged not called after VE.Clear().")
		End Sub
#Else
	  <TestMethod, Tag("Model"), Tag("Mods")>
	  Public Sub ShouldNotHaveErrorsAfterRejectChanges()
		  Dim hasError = WhenModifyingEntities.HasError(TestCustomer, "CompanyName")
		  Assert.IsFalse(hasError(), "Should not have errors at the start")

		  TestCustomer.CompanyName = Nothing
		  Assert.IsTrue(hasError(), "Should have error after nulling name")

		  TestCustomer.EntityAspect.RejectChanges()
		  TestCustomer.EntityAspect.ValidationErrors.Clear() ' ToDo: DevForce should do this?
		  Assert.IsFalse(hasError(), "Should not have errors after reject changes")
	  End Sub
	  Private Shared Function HasError(ByVal entity As Entity, ByVal propertyName As String) As Func(Of Boolean)
		  Dim idei = CType(entity, System.ComponentModel.IDataErrorInfo)
		  Return Function() Not String.IsNullOrEmpty(idei(propertyName))
	  End Function
#End If
  End Class

  <TestClass>
  Public Class CacheOnlyContext
  ' Silverlight Test Framework won't let us make abstract (as we might wish) 

	<TestInitialize>
	Public Sub TestInitialize()
	  Manager = PrepareNewOfflineManager()
	  Context()
	End Sub

	Protected Overridable Sub Context()
	End Sub


	' Create manager that is disconnected and primed with test entities
	Protected Function PrepareNewOfflineManager() As NorthwindManager

	  ' don't connect. Do everything in cache
	  Dim manager = New NorthwindManager(False)
	  manager.BlockFetching()

	  ' create some test entities
	  TestCustomer = New Customer With {.CustomerID = TestCustomerId, .CompanyName = "Test Company", .Country = "USA"}
	  TestEmployee = New Employee With {.EmployeeID = TestEmployeeId, .FirstName = "Joe", .LastName = "Blow"}

	  ' attach to the manager as if they previously existed and are unmodified.
	  manager.AttachEntity(TestCustomer)
	  manager.AttachEntity(TestEmployee)
	  manager.AttachEntities(MakeOrders(10, TestCustomerId, TestEmployeeId))
	  manager.AttachEntities(MakeMoreCustomers())

	  Return manager
	End Function

	Protected Property TestCustomer() As Customer
	Protected Property TestEmployee() As Employee

	Private Shared Function MakeOrders(ByVal howMany As Integer, ByVal custId? As Guid, ByVal empId As Integer) As IEnumerable(Of Order)
		Dim orderDates = New Date?() { New Date(1996, 1, 1), New Date(1997, 1, 1), New Date(1998, 1, 1), Nothing }

		Dim orderID = 1
		Dim tempVar As boolean = howMany > 0
      howMany -= 1
      Dim orderlist As New List(Of Order)
		Do While tempVar
'INSTANT VB TODO TASK: VB does not support iterators and has no equivalent to the C# 'yield' keyword:
        'yield Return New Order With {.OrderID = orderID, .CustomerID = custId, .EmployeeID = empId, .OrderDate = orderDates(howMany Mod 4)}

        orderlist.Add(New Order With {.OrderID = orderID, .CustomerID = custId, .EmployeeID = empId, .OrderDate = orderDates(howMany Mod 4)})
        orderID += 1
			tempVar = howMany > 0
			howMany -= 1
      Loop
      Return orderlist
	End Function

	Private Shared Function MakeMoreCustomers() As IEnumerable(Of Customer)
'INSTANT VB TODO TASK: VB does not support iterators and has no equivalent to the C# 'yield' keyword:
      '		yield Return New Customer With {.CustomerID = Guid.NewGuid(), .CompanyName = "UK1 Test Company", .Country = "UK"}
      ''INSTANT VB TODO TASK: VB does not support iterators and has no equivalent to the C# 'yield' keyword:
      '		yield Return New Customer With {.CustomerID = Guid.NewGuid(), .CompanyName = "UK2 Test Company", .Country = "UK"}
      ''INSTANT VB TODO TASK: VB does not support iterators and has no equivalent to the C# 'yield' keyword:
      '		yield Return New Customer With {.CustomerID = Guid.NewGuid(), .CompanyName = "French1 Test Company", .Country = "France"}
      ''INSTANT VB TODO TASK: VB does not support iterators and has no equivalent to the C# 'yield' keyword:
      '		yield Return New Customer With {.CustomerID = Guid.NewGuid(), .CompanyName = "French2 Test Company", .Country = "France"}
      ''INSTANT VB TODO TASK: VB does not support iterators and has no equivalent to the C# 'yield' keyword:
      '		yield Return New Customer With {.CustomerID = Guid.NewGuid(), .CompanyName = "USA2 Test Company", .Country = "USA"}
      ''INSTANT VB TODO TASK: VB does not support iterators and has no equivalent to the C# 'yield' keyword:
      '		yield Return New Customer With {.CustomerID = Guid.NewGuid(), .CompanyName = "USA3 Test Company", .Country = "USA"}

      Dim custlist As New List(Of Customer)
      custlist.Add(New Customer With {.CustomerID = Guid.NewGuid(), .CompanyName = "UK1 Test Company", .Country = "UK"})
      custlist.Add(New Customer With {.CustomerID = Guid.NewGuid(), .CompanyName = "UK2 Test Company", .Country = "UK"})
      custlist.Add(New Customer With {.CustomerID = Guid.NewGuid(), .CompanyName = "French1 Test Company", .Country = "France"})
      custlist.Add(New Customer With {.CustomerID = Guid.NewGuid(), .CompanyName = "French2 Test Company", .Country = "France"})
      custlist.Add(New Customer With {.CustomerID = Guid.NewGuid(), .CompanyName = "USA2 Test Company", .Country = "USA"})
      custlist.Add(New Customer With {.CustomerID = Guid.NewGuid(), .CompanyName = "USA3 Test Company", .Country = "USA"})
      Return custlist
	End Function

	Protected Property Manager() As NorthwindManager
	Protected Shared ReadOnly TestCustomerId As Guid = Guid.NewGuid()
	Protected Const TestEmployeeId As Integer = 42

	Protected Sub ShouldGetCustomersUsingQuery(ByVal query As IEntityQuery(Of Customer))
	  Dim custs = Manager.ExecuteQuery(query)
	  custs.ShouldNotBeEmpty("Can't find customers in cache")
	End Sub

	Protected Sub ShouldGetEmployeesUsingQuery(ByVal query As IEntityQuery(Of Employee))
	  Dim emps = Manager.ExecuteQuery(query)
	  emps.ShouldNotBeEmpty("Can't find Employees in cache")
	End Sub

  End Class
End Namespace
